import os
import queue
import time
from multiprocessing import Array, Pool, Process, Queue, cpu_count
from random import shuffle
from sys import stdin
from typing import Any, List, Tuple

import numpy as np
from pydantic import BaseModel
from wasabi import msg

from ...agents.episode_memory import rnn_weighted_history
from ...agents.mcts import MCTS
from ...envs.gym.mathy_gym_env import MathyGymEnv
from ...state import (
    MathyEnvState,
    MathyObservation,
    MathyWindowObservation,
    observations_to_window,
)
from ...util import discount, is_terminal_transition, pad_array, print_error
from ..policy_value_model import PolicyValueModel
from .config import SelfPlayConfig
from .trainer import SelfPlayTrainer
from .types import EpisodeHistory, EpisodeSummary


class PracticeRunner:
    """
    Instance that controls how episodes are executed. By default this class executes episodes serially
    in a single process. This is great for debugging problems in an interactive debugger or running locally
    but is not ideal for machines with many processors available. For multiprocessing swap out the default 
    `PracticeRunner` class for the `ParallelPracticeRunner` class that is defined below.    
    """

    def __init__(self, config: SelfPlayConfig):
        if config is None or not isinstance(config, SelfPlayConfig):
            raise ValueError(
                f"configuration must be an instance of {type(SelfPlayConfig)}"
            )
        self.config = config

    def get_env(self):
        raise NotImplementedError("game implementation must be provided by subclass")

    def get_model(self, game) -> PolicyValueModel:
        raise NotImplementedError(
            "predictor implementation must be provided by subclass"
        )

    def step(
        self,
        game: MathyGymEnv,
        env_state: MathyEnvState,
        mcts: MCTS,
        model: PolicyValueModel,
        move_count,
        history: List[EpisodeHistory],
        is_verbose_worker: bool,
    ):
        import tensorflow as tf

        # Hold on to the episode example data for training the neural net
        valids = game.mathy.get_valid_moves(env_state)
        rnn_state_h = tf.squeeze(model.embedding.state_h).numpy().tolist()
        rnn_state_c = tf.squeeze(model.embedding.state_c).numpy().tolist()
        rnn_history_h = (
            tf.squeeze(
                rnn_weighted_history(
                    [o.observation for o in history], len(rnn_state_h)
                )[0]
            )
            .numpy()
            .tolist()
        )
        last_observation: MathyObservation = env_state.to_observation(
            valids,
            rnn_state_h=rnn_state_h,
            rnn_state_c=rnn_state_c,
            rnn_history_h=rnn_history_h,
        )

        # If the move_count is less than threshold, set temp = 1 else 0
        temp = int(
            move_count < self.config.temperature_threshold * game.mathy.max_moves
        )
        mcts_state_copy = env_state.clone()
        predicted_policy, value = mcts.estimate_policy(
            mcts_state_copy, [rnn_state_h, rnn_state_c], temp=temp
        )
        action = np.random.choice(len(predicted_policy), p=predicted_policy)
        observation, reward, done, meta = game.step(action)
        next_state = game.state
        assert next_state is not None
        transition = meta["transition"]
        if is_verbose_worker and self.config.print_training is True:
            game.render()
        example_text = next_state.agent.problem
        r = float(transition.reward)
        is_term = is_terminal_transition(transition)
        is_win = True if is_term and r > 0 else False
        history.append(
            EpisodeHistory(
                text=example_text,
                action=action,
                reward=r,
                # NOTE: we have to update this when the episode is complete
                discounted=r,
                terminal=is_term,
                observation=last_observation,
                pi=predicted_policy,
                value=float(value),
            )
        )

        # Keep going if the reward signal is not terminal
        if not is_term:
            return next_state, None

        rewards = [x.reward for x in history]
        discounts = list(discount(rewards))
        episode_reward = np.sum(discounts)
        # Build a final history with discounted rewards set
        final_history: List[EpisodeHistory] = []
        for i, h in enumerate(history):
            final_history.append(
                EpisodeHistory(
                    text=h.text,
                    action=h.action,
                    reward=h.reward,
                    discounted=float(discounts[i]),
                    terminal=h.terminal,
                    observation=h.observation,
                    pi=h.pi,
                    value=h.value,
                )
            )

        return next_state, (final_history, episode_reward, is_win)

    def execute_episodes(
        self, episode_args_list
    ) -> Tuple[List[EpisodeHistory], List[EpisodeSummary]]:
        """
        Execute (n) episodes of self-play serially. This is mostly useful for debugging, and
        when you cannot fit multiple copies of your model in GPU memory
        """
        examples = []
        results: List[EpisodeSummary] = []
        if self.config.profile:
            import cProfile

            pr = cProfile.Profile()
            pr.enable()

        try:
            game = self.get_env()
            predictor = self.get_model(game)
            for i, args in enumerate(episode_args_list):
                start = time.time()
                (
                    episode_examples,
                    episode_reward,
                    is_win,
                    problem,
                ) = self.execute_episode(i, game, predictor, **args)
                duration = time.time() - start
                examples.extend(episode_examples)
                episode_summary = EpisodeSummary(
                    solved=bool(is_win),
                    text=problem.text,
                    complexity=problem.complexity,
                    reward=episode_reward,
                    duration=duration,
                )
                results.append(episode_summary)
                self.episode_complete(i, episode_summary)
        except KeyboardInterrupt:
            print("Interrupt received. Exiting.")

        if self.config.profile:
            profile_name = f"worker_0.profile"
            profile_path = os.path.join(self.config.model_dir, profile_name)
            pr.disable()
            pr.dump_stats(profile_path)
            if self.config.verbose:
                print(f"PROFILER: saved {profile_path}")
        return examples, results

    def execute_episode(
        self,
        episode: int,
        game: MathyGymEnv,
        predictor: PolicyValueModel,
        model_dir: str,
        is_verbose_worker: bool = False,
    ):
        """
        This function executes one episode.
        As the game is played, each turn is added as a training example to
        trainExamples. The game continues until get_state_reward returns a non-zero
        value, then the outcome of the game is used to assign values to each example
        in trainExamples.
        """
        if game is None:
            raise ValueError("PracticeRunner.get_env returned None type")
        if predictor is None:
            raise ValueError("PracticeRunner.get_model returned None type")
        game.reset()
        if game.state is None:
            raise ValueError("Cannot start self-play practice with a None game state.")
        env_state = game.state
        episode_history: List[Any] = []
        move_count = 0
        mcts = MCTS(game.mathy, predictor, self.config.cpuct, self.config.mcts_sims)
        if is_verbose_worker and self.config.print_training is True:
            game.render()

        while True:
            move_count += 1
            env_state, result = self.step(
                game=game,
                env_state=env_state,
                mcts=mcts,
                model=predictor,
                move_count=move_count,
                history=episode_history,
                is_verbose_worker=is_verbose_worker,
            )
            if result is not None:
                if is_verbose_worker and self.config.print_training is True:
                    game.render()

                return result + (game.problem,)

    def episode_complete(self, episode: int, summary: EpisodeSummary):
        """Called after each episode completes. Useful for things like updating
        progress indicators"""
        pass

    def process_trained_model(
        self, updated_model, iteration, train_examples, model_path
    ):
        if updated_model is None:
            return False
        # updated_model.save_checkpoint(model_path)
        return True

    def train(self, iteration, train_examples, model_path=None):
        """
        Train the model at the given checkpoint path with the training examples and
        return the updated model or None if there was an error."""
        return self.process_trained_model(
            self.train_with_examples(iteration, train_examples, model_path),
            iteration,
            train_examples,
            model_path,
        )

    def train_with_examples(self, iteration, train_examples, model_path=None):
        game = self.get_env()
        new_net = self.get_model(game)
        trainer = SelfPlayTrainer(
            self.config, new_net, action_size=new_net.predictions
        )
        if trainer.train(train_examples, new_net):
            new_net.save()


class ParallelPracticeRunner(PracticeRunner):
    """Run (n) parallel self-play or training processes in parallel."""

    request_quit = False

    def execute_episodes(
        self, episode_args_list
    ) -> Tuple[List[EpisodeHistory], List[EpisodeSummary]]:
        def worker(worker_idx: int, work_queue: Queue, result_queue: Queue):
            """Pull items out of the work queue and execute episodes until there are
            no items left """
            game = self.get_env()
            predictor = self.get_model(game)
            msg.good(f"Worker {worker_idx} started.")

            while (
                ParallelPracticeRunner.request_quit is False
                and work_queue.empty() is False
            ):
                episode, args = work_queue.get()
                start = time.time()
                try:
                    (
                        episode_examples,
                        episode_reward,
                        is_win,
                        problem,
                    ) = self.execute_episode(
                        episode,
                        game,
                        predictor,
                        is_verbose_worker=worker_idx == 0,
                        **args,
                    )
                except KeyboardInterrupt:
                    break
                except Exception as e:
                    err = print_error(e, f"Self-practice episode threw")
                    result_queue.put((i, [], {"error": err}))
                    continue
                duration = time.time() - start
                episode_summary = EpisodeSummary(
                    complexity=problem.complexity,
                    text=problem.text,
                    reward=episode_reward,
                    solved=bool(is_win),
                    duration=duration,
                )
                result_queue.put((i, episode_examples, episode_summary))
            return 0

        # Fill a work queue with episodes to be executed.
        work_queue: Queue = Queue()
        result_queue: Queue = Queue()
        for i, args in enumerate(episode_args_list):
            work_queue.put((i, args))
        processes = [
            Process(target=worker, args=(i, work_queue, result_queue), daemon=True)
            for i in range(self.config.num_workers)
        ]
        for proc in processes:
            proc.start()

        # Gather the outputs
        examples = []
        results = []
        count = 0
        while ParallelPracticeRunner.request_quit is False and count != len(
            episode_args_list
        ):
            try:
                i, episode_examples, summary = result_queue.get_nowait()
                self.episode_complete(i, summary)
                count += 1
                if "error" not in summary:
                    examples.extend(episode_examples)
                    results.append(summary)
            except queue.Empty:
                pass

        # Wait for the workers to exit completely
        for proc in processes:
            proc.join()

        return examples, results

    def train(self, iteration, train_examples, model_path=None):
        def train_and_save(output, i, examples, out_path):
            update_model = self.train_with_examples(i, examples, out_path)
            output.put(self.process_trained_model(update_model, i, examples, out_path))

        result_queue = Queue()
        proc = Process(
            target=train_and_save,
            args=(result_queue, iteration, train_examples, model_path),
        )
        proc.start()
        result = result_queue.get()
        proc.join()
        return result



from json import JSONEncoder
from typing import Any, List
import json
import os
import sys
import tempfile
import time
import types
from collections import deque
from multiprocessing import Array, Pool, Process, Queue, Value, cpu_count
from pathlib import Path
from pickle import Pickler, Unpickler
from random import shuffle
from shutil import copyfile
from .config import SelfPlayConfig
import numpy as np

import srsly
from .types import EpisodeHistory, EpisodeSummary
from ...agents.mcts import MCTS
from .lib.progress.bar import Bar
from .lib.average_meter import AverageMeter

INPUT_EXAMPLES_FILE_NAME = "examples.jsonl"


class PracticeSession:
    """
    This class executes the self-play + learning. It uses the functions defined
    in Game and NeuralNet. args are specified in main.py.
    """

    all_examples: List[EpisodeHistory] = []

    def has_examples(self) -> bool:
        return bool(len(self.all_examples) > 0)

    def __init__(self, runner, config: SelfPlayConfig, env_name: str):
        self.runner = runner
        self.config = config
        self.env = env_name
        self.problem_count = self.runner.config.self_play_problems
        self.all_examples = []
        self.skip_first_self_play = False
        loaded = self.load_training_examples()
        if loaded is not False:
            print("Loaded examples from: {}".format(loaded))

    def learn(self):
        iterations = self.problem_count
        for i in range(1, self.config.training_iterations + 1):
            solve, fail, summary, new_examples = self.run_self_play(i, iterations)
            self.run_network_training(i)

    def run_self_play(self, iteration, num_episodes):
        bar = Bar(self.env, max=num_episodes)
        bar.suffix = "working on first problem..."
        bar.next()
        current_episode = 0

        solved = 0
        failed = 0

        def episode_complete(self, episode, summary):
            nonlocal current_episode, bar, num_episodes, solved, failed
            current_episode += 1
            if summary.solved is True:
                solved = solved + 1
            else:
                failed = failed + 1

            bar.message = "{} [{},{}]".format(self.env.upper(), solved, failed)
            bar.suffix = "({eps}/{maxeps}) | total: {total:} | remaining: {eta:}".format(
                eps=current_episode,
                maxeps=num_episodes,
                total=bar.elapsed_td,
                eta=bar.eta_td,
            )
            bar.next()

        episodes_with_args = []
        for i in range(1, num_episodes + 1):
            episodes_with_args.append(dict(model_dir=self.runner.config.model_dir))

        old_update = self.runner.episode_complete
        self.runner.episode_complete = types.MethodType(episode_complete, self)
        new_examples, episodes_rewards = self.runner.execute_episodes(
            episodes_with_args
        )
        # Output a few solve/fail stats
        complexity_stats = dict()
        solve = 0
        fail = 0

        def add_stat(complexity, value):
            nonlocal complexity_stats, solve, fail
            if complexity not in complexity_stats:
                complexity_stats[complexity] = dict(solve=0, fail=0)
            if value >= 0:
                solve += 1
                complexity_stats[complexity]["solve"] += 1
            elif value < 0:
                fail += 1
                complexity_stats[complexity]["fail"] += 1

        for summary in episodes_rewards:
            value = summary.reward
            complexity = str(summary.complexity)
            add_stat(complexity, value)
        print(
            "\n\nPractice results:\n --- Solved ({}) --- Failed ({})".format(
                solve, fail
            )
        )
        print("By complexity:\n{}\n\n".format(json.dumps(complexity_stats, indent=2)))
        self.runner.episode_complete = old_update
        bar.finish()
        self.all_examples.extend(new_examples)
        self.save_training_examples()
        return solve, fail, summary, new_examples

    def run_network_training(self, iteration):
        """
        Train the current best model with the gathered training examples and return
        a tuple of (best, new) where best is the existing best trained model (or a blank
        one) and new is the model that was just trained.
        """
        train_examples = self.all_examples
        # print(train_examples)
        print("Training with {} examples".format(len(train_examples)))
        return self.runner.train(
            iteration, train_examples, self.runner.config.model_dir
        )

    def load_training_examples(self):
        file_path = Path(self.runner.config.model_dir or "") / INPUT_EXAMPLES_FILE_NAME
        if not file_path.is_file():
            return False
        examples = list(srsly.read_jsonl(str(file_path)))
        # with file_path.open("r", encoding="utf8") as f:
        #     for line in f:
        #         ex = ujson.loads(line)
        #         examples.append(ex)
        self.all_examples = examples
        self.skip_first_self_play = True
        return str(file_path)

    def save_training_examples(self):
        model_dir = Path(self.runner.config.model_dir)
        if not model_dir.is_dir():
            model_dir.mkdir(parents=True, exist_ok=True)

        # Write to local then copy (don't thrash virtual file systems like GCS)
        _, tmp_file = tempfile.mkstemp()
        srsly.write_jsonl(tmp_file, self.all_examples)
        out_file = model_dir / INPUT_EXAMPLES_FILE_NAME
        copyfile(tmp_file, str(out_file))
        os.remove(tmp_file)
        return str(out_file)



from gym import spaces
import numpy as np


class MaskedDiscrete(spaces.Discrete):
    r"""A masked discrete space in :math:`\{ 0, 1, \\dots, n-1 \}`.
    Example::
        >>> MaskedDiscrete(3, mask=(1,1,0))
    """
    mask: np.array

    def __init__(self, n, mask):
        assert isinstance(mask, (tuple, list))
        assert len(mask) == n
        self.mask = np.array(mask)
        super(MaskedDiscrete, self).__init__(n)

    def sample(self):
        probability = self.mask / np.sum(self.mask)
        return self.np_random.choice(self.n, p=probability)



from typing import Any, Dict, List, Optional, Tuple, Union

import tensorflow as tf

from mathy.agents.base_config import BaseConfig
from mathy.core.expressions import MathTypeKeysMax
from mathy.state import (
    MathyInputsType,
    MathyWindowObservation,
    ObservationFeatureIndices,
)
from mathy.agents.densenet import DenseNetStack


class MathyEmbedding(tf.keras.Model):
    def __init__(
        self,
        config: BaseConfig,
        episode_reset_state_h: Optional[bool] = True,
        episode_reset_state_c: Optional[bool] = True,
        **kwargs,
    ):
        super(MathyEmbedding, self).__init__(**kwargs)
        self.config = config
        self.episode_reset_state_h = episode_reset_state_h
        self.episode_reset_state_c = episode_reset_state_c
        self.init_rnn_state()
        self.token_embedding = tf.keras.layers.Embedding(
            input_dim=MathTypeKeysMax,
            output_dim=self.config.embedding_units,
            name="nodes_embedding",
            mask_zero=True,
        )

        # +1 for the value
        # +1 for the time
        # +2 for the problem type hashes
        self.concat_size = (
            4
            if self.config.use_env_features
            else 1
            if self.config.use_node_values
            else 0
        )
        # if self.config.use_node_values:
        #     self.values_dense = tf.keras.layers.Dense(self.config.units, name="values_input")
        if self.config.use_env_features:
            self.time_dense = tf.keras.layers.Dense(
                self.config.units, name="time_input"
            )
            self.type_dense = tf.keras.layers.Dense(
                self.config.units, name="type_input"
            )
        self.in_dense = tf.keras.layers.Dense(
            self.config.units,
            # In transform gets the embeddings concatenated with the
            # floating point value at each node.
            input_shape=(None, self.config.embedding_units + self.concat_size),
            name="in_dense",
            activation="relu",
            kernel_initializer="he_normal",
        )
        self.output_dense = tf.keras.layers.Dense(
            self.config.embedding_units,
            name="out_dense",
            activation="relu",
            kernel_initializer="he_normal",
        )

        NormalizeClass = tf.keras.layers.LayerNormalization
        if self.config.normalization_style == "batch":
            NormalizeClass = tf.keras.layers.BatchNormalization
        self.out_dense_norm = NormalizeClass(name="out_dense_norm")
        if self.config.use_lstm:
            self.nodes_lstm_norm = NormalizeClass(name="lstm_nodes_norm")
            self.lstm = tf.keras.layers.LSTM(
                self.config.lstm_units,
                name="lstm",
                return_sequences=True,
                time_major=False,
                return_state=True,
            )
        else:
            self.densenet = DenseNetStack(
                units=self.config.units,
                num_layers=6,
                output_transform=self.output_dense,
                normalization_style=self.config.normalization_style,
            )
            self.dense_attention = tf.keras.layers.Attention()

    def compute_output_shape(self, input_shapes: List[tf.TensorShape]) -> Any:
        nodes_shape: tf.TensorShape = input_shapes[0]
        return tf.TensorShape(
            (
                nodes_shape.dims[0].value,
                nodes_shape.dims[1].value,
                self.config.embedding_units,
            )
        )

    def init_rnn_state(self):
        """Track RNN states with variables in the graph"""
        self.state_c = tf.Variable(
            tf.zeros([1, self.config.lstm_units]),
            trainable=False,
            name="embedding/rnn/agent_state_c",
        )
        self.state_h = tf.Variable(
            tf.zeros([1, self.config.lstm_units]),
            trainable=False,
            name="embedding/rnn/agent_state_h",
        )

    def reset_rnn_state(self, force: bool = False):
        """Zero out the RNN state for a new episode"""
        if self.episode_reset_state_h or force is True:
            self.state_h.assign(tf.zeros([1, self.config.lstm_units]))

        if self.episode_reset_state_c or force is True:
            self.state_c.assign(tf.zeros([1, self.config.lstm_units]))

    def call(self, features: MathyInputsType, train: tf.Tensor = None) -> tf.Tensor:
        nodes = features[ObservationFeatureIndices.nodes]
        values = features[ObservationFeatureIndices.values]
        type = features[ObservationFeatureIndices.type]
        time = features[ObservationFeatureIndices.time]
        nodes_shape = tf.shape(features[ObservationFeatureIndices.nodes])
        batch_size = nodes_shape[0]  # noqa
        sequence_length = nodes_shape[1]

        in_rnn_state_h = features[ObservationFeatureIndices.rnn_state_h]
        in_rnn_state_c = features[ObservationFeatureIndices.rnn_state_c]
        in_rnn_history_h = features[ObservationFeatureIndices.rnn_history_h]

        with tf.name_scope("prepare_inputs"):
            values = tf.expand_dims(values, axis=-1, name="values_input")
            query = self.token_embedding(nodes)

            # if self.config.use_node_values:
            #     values = self.values_dense(values)

            # If not using env features, only concatenate the tokens and values
            if not self.config.use_env_features and self.config.use_node_values:
                query = tf.concat([query, values], axis=-1, name="tokens_and_values")
                query.set_shape(
                    (None, None, self.config.embedding_units + self.concat_size)
                )
            elif self.config.use_env_features:
                env_inputs = [
                    query,
                    self.type_dense(type),
                    self.time_dense(time),
                ]
                if self.config.use_node_values:
                    env_inputs.insert(1, values)

                query = tf.concat(env_inputs, axis=-1, name="tokens_and_values",)

        # Input dense transforms
        query = self.in_dense(query)

        if self.config.use_lstm:
            with tf.name_scope("rnn"):
                query, state_h, state_c = self.lstm(
                    query, initial_state=[in_rnn_state_h, in_rnn_state_c]
                )
                query = self.nodes_lstm_norm(query)

            self.state_h.assign(state_h[-1:])
            self.state_c.assign(state_c[-1:])

            # Concatenate the RNN output state with our historical RNN state
            # See: https://arxiv.org/pdf/1810.04437.pdf
            with tf.name_scope("combine_outputs"):
                rnn_state_with_history = tf.concat(
                    [state_h[-1:], in_rnn_history_h[-1:]], axis=-1,
                )
                lstm_context = tf.tile(
                    tf.expand_dims(rnn_state_with_history, axis=0),
                    [batch_size, sequence_length, 1],
                )
                # Combine the output LSTM states with the historical average LSTM states
                # and concatenate it with the query
                output = tf.concat(
                    [query, lstm_context], axis=-1, name="combined_outputs"
                )
                output = self.output_dense(output)
        else:
            # Non-recurrent model
            output = self.densenet(query)
            output = self.dense_attention([output, output])

        return self.out_dense_norm(output)



from typing import Any, Dict, List, Optional, Type

from ..env import MathyEnvProblem
from ..problems import gen_simplify_multiple_terms
from ..rules import (
    AssociativeSwapRule,
    ConstantsSimplifyRule,
    VariableMultiplyRule,
)
from ..core.rule import BaseRule
from ..state import MathyEnvState
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .poly_simplify import PolySimplify
from numpy.random import randint, uniform


class ComplexSimplify(PolySimplify):
    """A Mathy environment for simplifying complex terms (e.g. 4x^3 * 7y) inside of
    expressions. The goal is to simplify the complex term within the allowed number
    of environment steps.
    """

    def get_env_namespace(self) -> str:
        return "mathy.monomials.complex_simplify"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity * 5

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Given a set of parameters to control term generation, produce
        a complex term that has a simple representation that must be found.
        - "4x * 2y^2 * 7q"
        - "7j * 2z^6"
        - "x * 2y^7 * 8z * 2x"
        """

        if params.difficulty == MathyEnvDifficulty.easy:
            num_terms = randint(2, 6)
            scaling = uniform(0.5, 0.9)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op="*",
                optional_var=True,
                inner_terms_scaling=scaling,
                powers_probability=uniform(0.2, 0.5),
                noise_probability=uniform(0.2, 0.5),
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            num_terms = randint(3, 6)
            scaling = uniform(0.5, 0.85)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op="*",
                optional_var=True,
                inner_terms_scaling=scaling,
                powers_probability=uniform(0.4, 0.7),
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            num_terms = randint(4, 8)
            scaling = uniform(0.2, 0.6)
            text, complexity = gen_simplify_multiple_terms(
                num_terms,
                op="*",
                shuffle_probability=uniform(0.3, 0.8),
                powers_probability=uniform(0.4, 0.9),
                inner_terms_scaling=scaling,
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, complexity, self.get_env_namespace())



from typing import Optional

from .agents.base_config import BaseConfig
from .agents.episode_memory import EpisodeMemory
from .agents.policy_value_model import PolicyValueModel, load_policy_value_model


class Mathy:
    """The standard interface for working with Mathy models and agents."""

    config: BaseConfig
    model: PolicyValueModel

    def __init__(
        self,
        *,
        model_path: str = None,
        model: PolicyValueModel = None,
        config: BaseConfig = None,
        silent: bool = False,
    ):
        if model_path is not None:
            self.model, self.config = load_policy_value_model(model_path, silent=silent)
        elif model is not None and config is not None:
            if not isinstance(model, PolicyValueModel):
                raise ValueError("model must derive PolicyValueModel for compatibility")
            self.model = model
            self.config = config
        else:
            raise ValueError(
                "Either 'model_path' or ('model' and 'config') must be provided"
            )

    def simplify(
        self,
        *,
        model: str = "mathy_alpha_sm",
        problem: str,
        max_steps: int = 20,
        thinking_steps: int = 3,
    ) -> EpisodeMemory:
        """Simplify an input problem using the PolySimplify environment.
        
        # Arguments
        model (str): The input model to use for picking simplifying actions
        problem (str): The ascii math problem text, e.g. `-(4 + 2x) * 8 / 7y^(3 - 2)`
        max_steps (int): The maximum number of episode steps to allow the agent to take
            while solving the problem. Taking more than this is considered a failure.
        thinking_steps (int): The number of timesteps to look at the problem before attempting
            to solve it. These steps **do not** count toward the `max_steps` argument total.

        # Returns
        (EpisodeMemory): The stored episode memory containing the intermediate steps to get
            to the solution for the input problem.

        """
        import gym
        import tensorflow as tf
        from colr import color
        from .envs.gym import MathyGymEnv
        from .agents.action_selectors import A3CEpsilonGreedyActionSelector
        from .state import observations_to_window, MathyObservation
        from .agents.policy_value_model import PolicyValueModel
        from .util import calculate_grouping_control_signal

        environment = "poly"
        difficulty = "easy"
        episode_memory = EpisodeMemory()
        env: MathyGymEnv = gym.make(f"mathy-{environment}-{difficulty}-v0")
        last_observation: MathyObservation = env.reset_with_input(
            problem_text=problem, rnn_size=self.config.lstm_units, max_moves=max_steps
        )
        assert env.state is not None
        last_text = env.state.agent.problem
        last_action = -1
        last_reward = -1

        selector = A3CEpsilonGreedyActionSelector(
            model=self.model, episode=0, worker_id=0, epsilon=0
        )

        # Set RNN to 0 state for start of episode
        selector.model.embedding.reset_rnn_state()

        # Start with the "init" sequence [n] times
        for i in range(self.config.num_thinking_steps_begin + 1):
            rnn_state_h = tf.squeeze(selector.model.embedding.state_h.numpy())
            rnn_state_c = tf.squeeze(selector.model.embedding.state_c.numpy())
            seq_start = env.state.to_start_observation(rnn_state_h, rnn_state_c)
            selector.model(observations_to_window([seq_start]).to_inputs())

        done = False
        while not done:
            env.render(self.config.print_mode, None)
            # store rnn state for replay training
            rnn_state_h = tf.squeeze(selector.model.embedding.state_h.numpy())
            rnn_state_c = tf.squeeze(selector.model.embedding.state_c.numpy())
            last_rnn_state = [rnn_state_h, rnn_state_c]

            # named tuples are read-only, so add rnn state to a new copy
            last_observation = MathyObservation(
                nodes=last_observation.nodes,
                mask=last_observation.mask,
                values=last_observation.values,
                type=last_observation.type,
                time=last_observation.time,
                rnn_state_h=rnn_state_h,
                rnn_state_c=rnn_state_c,
                rnn_history_h=episode_memory.rnn_weighted_history(
                    self.config.lstm_units
                )[0],
            )
            window = episode_memory.to_window_observation(last_observation)
            try:
                action, value = selector.select(
                    last_state=env.state,
                    last_window=window,
                    last_action=last_action,
                    last_reward=last_reward,
                    last_rnn_state=last_rnn_state,
                )
            except KeyboardInterrupt:
                print("Done!")
                return episode_memory
            except BaseException as e:
                print("Prediction failed with error:", e)
                print("Inputs to model are:", window)
                continue
            # Take an env step
            observation, reward, done, _ = env.step(action)
            rnn_state_h = tf.squeeze(selector.model.embedding.state_h.numpy())
            rnn_state_c = tf.squeeze(selector.model.embedding.state_c.numpy())
            observation = MathyObservation(
                nodes=observation.nodes,
                mask=observation.mask,
                values=observation.values,
                type=observation.type,
                time=observation.time,
                rnn_state_h=rnn_state_h,
                rnn_state_c=rnn_state_c,
                rnn_history_h=episode_memory.rnn_weighted_history(
                    self.config.lstm_units
                )[0],
            )

            new_text = env.state.agent.problem
            grouping_change = calculate_grouping_control_signal(
                last_text, new_text, clip_at_zero=self.config.clip_grouping_control
            )
            episode_memory.store(
                observation=last_observation,
                action=action,
                reward=reward,
                grouping_change=grouping_change,
                value=value,
            )
            if done:
                # Last timestep reward
                win = reward > 0.0
                env.render(self.config.print_mode, None)
                print(
                    color(
                        text="SOLVE" if win else "FAIL", fore="green" if win else "red",
                    )
                )
                break

            last_observation = observation
            last_action = action
            last_reward = reward
        return episode_memory



from typing import Any, List, Optional, Type

from numpy.random import randint, uniform

from .. import time_step
from ..core.expressions import MathExpression
from ..core.rule import BaseRule
from ..env import MathyEnv, MathyEnvProblem
from ..problems import (
    gen_binomial_times_binomial,
    gen_binomial_times_monomial,
    rand_bool,
)
from ..rules import (
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveMultiplyRule,
    VariableMultiplyRule,
)
from ..state import MathyEnvState, MathyObservation
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from ..util import get_terms, has_like_terms, is_preferred_term_form


class BinomialDistribute(MathyEnv):
    """A Mathy environment for distributing pairs of binomials.

    The FOIL method is sometimes used to solve these types of problems, where
    FOIL is just the distributive property applied to two binomials connected
    with a multiplication."""

    def get_env_namespace(self) -> str:
        return "mathy.binomials.mulptiply"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity * 4

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If there are no like terms."""
        if not has_like_terms(expression):
            term_nodes = get_terms(expression)
            is_win = True
            for term in term_nodes:
                if not is_preferred_term_form(term):
                    is_win = False
            if is_win:
                return time_step.termination(features, self.get_win_signal(env_state))
        return None

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Given a set of parameters to control term generation, produce
        2 binomials expressions connected by a multiplication. """
        if params.difficulty == MathyEnvDifficulty.easy:
            if rand_bool(50):
                text, complexity = gen_binomial_times_monomial(min_vars=2, max_vars=3)
            else:
                text, complexity = gen_binomial_times_binomial(
                    min_vars=2,
                    max_vars=3,
                    powers_probability=uniform(0.1, 0.4),
                    like_variables_probability=uniform(0.3, 0.7),
                )
        elif params.difficulty == MathyEnvDifficulty.normal:
            text, complexity = gen_binomial_times_binomial(
                min_vars=2,
                max_vars=2,
                powers_probability=uniform(0.2, 0.6),
                like_variables_probability=uniform(0.2, 0.5),
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            text, complexity = gen_binomial_times_binomial(
                min_vars=2,
                max_vars=3,
                simple_variables=False,
                powers_probability=uniform(0.4, 0.8),
                like_variables_probability=uniform(0.1, 0.3),
            )
            complexity += 2
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, complexity, self.get_env_namespace())



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TensorFlow ops for continuous-action Policy Gradient algorithms."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
from six.moves import zip
import tensorflow as tf
from . import base_ops
from . import value_ops

from tensorflow.python.util import nest

PolicyEntropyExtra = collections.namedtuple("policy_entropy_extra", ["entropy"])
SequenceA2CExtra = collections.namedtuple(
    "sequence_a2c_extra",
    [
        "entropy",
        "entropy_loss",
        "baseline_loss",
        "policy_gradient_loss",
        "advantages",
        "discounted_returns",
    ],
)


def policy_gradient(
    policies, actions, action_values, policy_vars=None, name="policy_gradient"
):
    """Computes policy gradient losses for a batch of trajectories.

  See `policy_gradient_loss` for more information on expected inputs and usage.

  Args:
    policies: A distribution over a batch supporting a `log_prob` method, e.g.
        an instance of `tfp.distributions.Distribution`. For example, for
        a diagonal gaussian policy:
        `policies = tfp.distributions.MultivariateNormalDiag(mus, sigmas)`
    actions: An action batch Tensor used as the argument for `log_prob`. Has
        shape equal to the batch shape of the policies concatenated with the
        event shape of the policies (which may be scalar, in which case
        concatenation leaves shape just equal to batch shape).
    action_values: A Tensor containing estimates of the values of the `actions`.
        Has shape equal to the batch shape of the policies.
    policy_vars: An optional iterable of Tensors used by `policies`. If provided
        is used in scope checks. For the multivariate normal example above this
        would be `[mus, sigmas]`.
    name: Customises the name_scope for this op.

  Returns:
    loss: Tensor with same shape as `actions` containing the total loss for each
        element in the batch. Differentiable w.r.t the variables in `policies`
        only.
  """
    policy_vars = list(policy_vars) if policy_vars else list()
    with tf.compat.v1.name_scope(
        values=policy_vars + [actions, action_values], name=name
    ):
        actions = tf.stop_gradient(actions)
        action_values = tf.stop_gradient(action_values)
        log_prob_actions = policies.log_prob(actions)
        # Prevent accidental broadcasting if possible at construction time.
        action_values.get_shape().assert_is_compatible_with(
            log_prob_actions.get_shape()
        )
        return -tf.multiply(log_prob_actions, action_values)


def policy_gradient_loss(
    policies, actions, action_values, policy_vars=None, name="policy_gradient_loss"
):
    """Computes policy gradient losses for a batch of trajectories.

  This wraps `policy_gradient` to accept a possibly nested array of `policies`
  and `actions` in order to allow for multiple action distribution types or
  independent multivariate distributions if not directly available. It also sums
  up losses along the time dimension, and is more restrictive about shapes,
  assuming a [T, B] layout for the `batch_shape` of the policies and a
  concatenate(`[T, B]`, `event_shape` of the policies) shape for the actions.

  Args:
    policies: A (possibly nested structure of) distribution(s) supporting
        `batch_shape` and `event_shape` properties along with a `log_prob`
        method (e.g. an instance of `tfp.distributions.Distribution`),
        with `batch_shape` equal to `[T, B]`.
    actions: A (possibly nested structure of) N-D Tensor(s) with shape
        `[T, B, ...]` where the final dimensions are the `event_shape` of the
        corresponding distribution in the nested structure (the shape can be
        just `[T, B]` if the `event_shape` is scalar).
    action_values: Tensor of shape `[T, B]` containing an estimate of the value
        of the selected `actions`.
    policy_vars: An optional (possibly nested structure of) iterable(s) of
        Tensors used by `policies`. If provided is used in scope checks.
    name: Customises the name_scope for this op.

  Returns:
    loss: Tensor of shape `[B]` containing the total loss for each sequence
    in the batch. Differentiable w.r.t `policy_logits` only.
  """
    actions = nest.flatten(actions)
    if policy_vars:
        policy_vars = nest.flatten_up_to(policies, policy_vars)
    else:
        policy_vars = [list()] * len(actions)
    policies = nest.flatten(policies)

    # Check happens after flatten so that we can be more flexible on nest
    # structures. This is equivalent to asserting that `len(policies) ==
    # len(actions)`, which is sufficient for what we're doing here.
    nest.assert_same_structure(policies, actions)

    for policies_, actions_ in zip(policies, actions):
        policies_.batch_shape.assert_has_rank(2)
        actions_.get_shape().assert_is_compatible_with(
            policies_.batch_shape.concatenate(policies_.event_shape)
        )

    scoped_values = policy_vars + actions + [action_values]
    with tf.compat.v1.name_scope(name, values=scoped_values):
        # Loss for the policy gradient. Doesn't push additional gradients through
        # the action_values.
        policy_gradient_loss_sequence = tf.add_n(
            [
                policy_gradient(policies_, actions_, action_values, pvars)
                for policies_, actions_, pvars in zip(policies, actions, policy_vars)
            ]
        )

        return tf.reduce_sum(
            input_tensor=policy_gradient_loss_sequence,
            axis=[0],
            name="policy_gradient_loss",
        )


def policy_entropy_loss(
    policies, policy_vars=None, scale_op=None, name="policy_entropy_loss"
):
    """Calculates entropy 'loss' for policies represented by a distributions.

  Given a (possible nested structure of) batch(es) of policies, this
  calculates the total entropy and corrects the sign so that minimizing the
  resulting loss op is equivalent to increasing entropy in the batch.

  This function accepts a nested structure of `policies` in order to allow for
  multiple distribution types or for multiple action dimensions in the case
  where there is no corresponding mutivariate form for available for a given
  univariate distribution. In this case, the loss is `sum_i(H(p_i, p_i))`
  where `p_i` are members of the `policies` nest. It can be shown that this is
  equivalent to calculating the entropy loss on the Cartesian product space
  over all the action dimensions, if the sampled actions are independent.

  The entropy loss is optionally scaled by some function of the policies.
  E.g. for Categorical distributions there exists such a scaling which maps
  the entropy loss into the range `[-1, 0]` in order to make it invariant to
  the size of the action space - specifically one can divide the loss by
  `sum_i(log(A_i))` where `A_i` is the number of categories in the i'th
  Categorical distribution in the `policies` nest).

  Args:
    policies: A (possibly nested structure of) batch distribution(s)
        supporting an `entropy` method that returns an N-D Tensor with shape
        equal to the `batch_shape` of the distribution, e.g. an instance of
        `tfp.distributions.Distribution`.
    policy_vars: An optional (possibly nested structure of) iterable(s) of
        Tensors used by `policies`. If provided is used in scope checks.
    scale_op: An optional op that takes `policies` as its only argument and
        returns a scalar Tensor that is used to scale the entropy loss.
        E.g. for Diag(sigma) Gaussian policies dividing by the number of
        dimensions makes entropy loss invariant to the action space dimension.
    name: Optional, name of this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B1, B2, ...]`.
    * `extra`: a namedtuple with fields:
        * `entropy`: entropy of the policy, shape `[B1, B2, ...]`.
    where [B1, B2, ... ] == policy.batch_shape
  """
    flat_policy_vars = nest.flatten(policy_vars) if policy_vars else list()
    with tf.compat.v1.name_scope(name, values=flat_policy_vars):
        # We want a value that we can minimize along with other losses, and where
        # minimizing means driving the policy towards a uniform distribution over
        # the actions. We thus scale it by negative one so that it can be simply
        # added to other losses.
        scale = tf.constant(-1.0, dtype=tf.float32)
        if scale_op:
            scale *= scale_op(policies)

        policies = nest.flatten(policies)
        entropy = tf.add_n([policy.entropy() for policy in policies], name="entropy")
        loss = tf.multiply(scale, entropy, name="entropy_loss")
        return base_ops.LossOutput(loss, PolicyEntropyExtra(entropy))


def sequence_a2c_loss(
    policies,
    baseline_values,
    actions,
    rewards,
    pcontinues,
    bootstrap_value,
    policy_vars=None,
    lambda_=1,
    entropy_cost=None,
    baseline_cost=1,
    entropy_scale_op=None,
    name="SequenceA2CLoss",
):
    """Constructs a TensorFlow graph computing the A2C/GAE loss for sequences.

  This loss jointly learns the policy and the baseline. Therefore, gradients
  for this loss flow through each tensor in `policies` and through each tensor
  in `baseline_values`, but no other input tensors. The policy is learnt with
  the advantage actor-critic loss, plus an optional entropy term. The baseline
  is regressed towards the n-step bootstrapped returns given by the
  reward/pcontinue sequence. The `baseline_cost` parameter scales the
  gradients w.r.t the baseline relative to the policy gradient, i.e.
  d(loss) / d(baseline) = baseline_cost * (n_step_return - baseline)`.

  This function is designed for batches of sequences of data. Tensors are
  assumed to be time major (i.e. the outermost dimension is time, the second
  outermost dimension is the batch dimension). We denote the sequence length in
  the shapes of the arguments with the variable `T`, the batch size with the
  variable `B`, neither of which needs to be known at construction time. Index
  `0` of the time dimension is assumed to be the start of the sequence.

  `rewards` and `pcontinues` are the sequences of data taken directly from the
  environment, possibly modulated by a discount. `baseline_values` are the
  sequences of (typically learnt) estimates of the values of the states
  visited along a batch of trajectories as observed by the agent given the
  sequences of one or more actions sampled from `policies`.

  The sequences in the tensors should be aligned such that an agent in a state
  with value `V` that takes an action `a` transitions into another state
  with value `V'`, receiving reward `r` and pcontinue `p`. Then `V`, `a`, `r`
  and `p` are all at the same index `i` in the corresponding tensors. `V'` is
  at index `i+1`, or in the `bootstrap_value` tensor if `i == T`.

  For n-dimensional action vectors, a multivariate distribution must be used
  for `policies`. In case there is no multivariate version for the desired
  univariate distribution, or in case the `actions` object is a nested
  structure (e.g. for multiple action types), this function also accepts a
  nested structure  of `policies`. In this case, the loss is given by
  `sum_i(loss(p_i, a_i))` where `p_i` are members of the `policies` nest, and
  `a_i` are members of the `actions` nest. We assume that a single baseline is
  used across all action dimensions for each timestep.

  Args:
    policies: A (possibly nested structure of) distribution(s) supporting
        `batch_shape` and `event_shape` properties & `log_prob` and `entropy`
        methods (e.g. an instance of `tfp.distributions.Distribution`),
        with `batch_shape` equal to `[T, B]`. E.g. for a (non-nested) diagonal
        multivariate gaussian with dimension `A` this would be:
        `policies = tfp.distributions.MultivariateNormalDiag(mus, sigmas)`
        where `mus` and `sigmas` have shape `[T, B, A]`.
    baseline_values: 2-D Tensor containing an estimate of the state value with
        shape `[T, B]`.
    actions: A (possibly nested structure of) N-D Tensor(s) with shape
        `[T, B, ...]` where the final dimensions are the `event_shape` of the
        corresponding distribution in the nested structure (the shape can be
        just `[T, B]` if the `event_shape` is scalar).
    rewards: 2-D Tensor with shape `[T, B]`.
    pcontinues: 2-D Tensor with shape `[T, B]`.
    bootstrap_value: 1-D Tensor with shape `[B]`.
    policy_vars: An optional (possibly nested structure of) iterables of
        Tensors used by `policies`. If provided is used in scope checks. For
        the multivariate normal example above this would be `[mus, sigmas]`.
    lambda_: an optional scalar or 2-D Tensor with shape `[T, B]` for
        Generalised Advantage Estimation as per
        https://arxiv.org/abs/1506.02438.
    entropy_cost: optional scalar cost that pushes the policy to have high
        entropy, larger values cause higher entropies.
    baseline_cost: scalar cost that scales the derivatives of the baseline
        relative to the policy gradient.
    entropy_scale_op: An optional op that takes `policies` as its only
        argument and returns a scalar Tensor that is used to scale the entropy
        loss. E.g. for Diag(sigma) Gaussian policies dividing by the number of
        dimensions makes entropy loss invariant to the action space dimension.
        See `policy_entropy_loss` for more info.
    name: Customises the name_scope for this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the total loss, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `entropy`: total loss per sequence, shape `[B]`.
        * `entropy_loss`: scaled entropy loss per sequence, shape `[B]`.
        * `baseline_loss`: scaled baseline loss per sequence, shape `[B]`.
        * `policy_gradient_loss`: policy gradient loss per sequence,
            shape `[B]`.
        * `advantages`: advantange estimates per timestep, shape `[T, B]`.
        * `discounted_returns`: discounted returns per timestep,
            shape `[T, B]`.
  """
    flat_policy_vars = nest.flatten(policy_vars) if policy_vars else list()
    scoped_values = (
        flat_policy_vars
        + nest.flatten(actions)
        + [baseline_values, rewards, pcontinues, bootstrap_value]
    )
    with tf.compat.v1.name_scope(name, values=scoped_values):
        # Loss for the baseline, summed over the time dimension.
        baseline_loss_td, td_lambda = value_ops.td_lambda(
            baseline_values, rewards, pcontinues, bootstrap_value, lambda_
        )

        # The TD error provides an estimate of the advantages of the actions.
        advantages = td_lambda.temporal_differences
        baseline_loss = tf.multiply(
            tf.convert_to_tensor(value=baseline_cost, dtype=tf.float32),
            baseline_loss_td,
            name="baseline_loss",
        )

        # Loss for the policy. Doesn't push additional gradients through
        # the advantages.
        pg_loss = policy_gradient_loss(
            policies, actions, advantages, policy_vars, name="policy_gradient_loss"
        )

        total_loss = tf.add(pg_loss, baseline_loss, name="total_loss")

        if entropy_cost is not None:
            loss, extra = policy_entropy_loss(policies, policy_vars, entropy_scale_op)
            entropy = tf.reduce_sum(
                input_tensor=extra.entropy, axis=0, name="entropy"
            )  # [B].
            entropy_loss = tf.multiply(
                tf.convert_to_tensor(value=entropy_cost, dtype=tf.float32),
                tf.reduce_sum(input_tensor=loss, axis=0),
                name="scaled_entropy_loss",
            )  # [B].
            total_loss = tf.add(
                total_loss, entropy_loss, name="total_loss_with_entropy"
            )
        else:
            entropy = None
            entropy_loss = None

        extra = SequenceA2CExtra(
            entropy=entropy,
            entropy_loss=entropy_loss,
            baseline_loss=baseline_loss,
            policy_gradient_loss=pg_loss,
            advantages=advantages,
            discounted_returns=td_lambda.discounted_returns,
        )
        return base_ops.LossOutput(total_loss, extra)




from ..mathy.core.tree import BinaryTreeNode, BinarySearchTree, STOP
from ..mathy.core.layout import TreeLayout, TreeMeasurement
import numpy as np
import pytest


def test_layout_tidier():
    tree = BinarySearchTree(0)
    for i in range(-10, 10):
        tree.insert(i)
    m: TreeMeasurement = TreeLayout().layout(tree)
    assert m is not None


def test_layout_tidier_aesthetic_one():
    """Verify layouts satisfy the "y-coordinate sharing" aesthetic.

    Build a tree, gather up all the nodes, group them by depth, and
    assert that each node in a depth group has the same y - coordinate"""
    tree = BinarySearchTree(0)
    for i in range(-5, 5):
        tree.insert(i)
    m: TreeMeasurement = TreeLayout().layout(tree)
    assert m is not None

    groups = {}

    def node_visit(node, depth, data):
        nonlocal groups
        if depth not in groups:
            groups[depth] = []
        groups[depth].append(node)

    tree.visit_preorder(node_visit)
    for k, v in groups.items():
        y_coords = [n.y for n in v]
        # converting to a set removes duplicates
        assert len(set(y_coords)) == 1


def test_layout_tidier_aesthetic_two():
    """Verify layouts satisfy the "children relative position" aesthetic.

    Build a tree, visit each node and assert that the left/right children are
    positioned to the left and right of its own coordinate.
    """
    tree = BinarySearchTree(0)
    for i in [-2, -1, -3, 2, 3, 1]:
        tree.insert(i)
    # Discard the measurement because each node has x/y attached
    TreeLayout().layout(tree)

    def node_visit(node: BinaryTreeNode, depth, data):
        if node.parent is None:
            return
        side = node.parent.get_side(node)
        if side == "left":
            # left child should be to the left of the parent (-x)
            assert node.x < node.parent.x
        elif side == "right":
            # left child should be to the left of the parent (+x)
            assert node.x > node.parent.x

    tree.visit_postorder(node_visit)


def test_layout_tidier_aesthetic_three():
    """Verify layouts satisfy the "equidistant children" aesthetic.

    Build a tree, visit each node and assert that its children are
    equidistant from it."""
    values = [-2, -1, -3, 2, 3, 1]
    tree = BinarySearchTree(0)
    for i in values:
        tree.insert(i)
    TreeLayout().layout(tree)

    def node_visit(node: BinaryTreeNode, depth: int, data):
        children = node.get_children()
        if len(children) != 2:
            return
        left = children[0].x
        right = children[1].x

        expected = left + abs(left - right) / 2
        # Parent node position matches expectation
        assert expected == node.x

    tree.visit_postorder(node_visit)


def test_layout_tidier_aesthetic_four_reflections():
    """Verify layouts satisfy the "subtree reflections" aesthetic.

    Build a tree with subtrees that are reflections of one another, and
    visit each node asserting that its mirror node is on the opposite
    side of the tree with the same distance to the root.
    """
    tree = BinarySearchTree(0)
    values = [-2, -1, -4, -3, -5, 2, 1, 4, 3, 5]
    for i in values:
        tree.insert(i)
    TreeLayout().layout(tree)
    for i in range(6):
        left = tree.find(-i)
        right = tree.find(i)
        left_dist = abs(left.x - tree.x)
        right_dist = abs(right.x - tree.x)
        assert left_dist == right_dist


def test_layout_tidier_aesthetic_four_identical_subtrees():
    """Verify layouts satisfy the "subtree identical regardless of position" aesthetic.

    Build a tree with known identical subtrees that are in different positions of the
    tree, and verify that their children have the same relative positions to them.
    """
    tree = BinarySearchTree(0)
    values = [7, 4, 3, 5, 13, 12, 14, -3, -6, -2, -7, -13, -14, -12]
    for i in values:
        tree.insert(i)
    TreeLayout().layout(tree)

    first_left_distance = None
    first_right_distance = None

    def node_visit(node, depth: int, data):
        nonlocal first_left_distance, first_right_distance
        # We only want to assert about the specific subtrees +-(13, 14, 12)
        if node.key != 13 and node.key != -13:
            return
        # expect lesser value to the left
        assert node.left.key == node.key - 1
        # expect greater value to the left
        assert node.right.key == node.key + 1

        less = node.left.x
        more = node.right.x
        left_dist = less - node.x
        right_dist = more - node.x
        if first_right_distance and first_left_distance:
            assert left_dist == first_left_distance
            assert right_dist == first_right_distance
        else:
            first_left_distance = left_dist
            first_right_distance = right_dist

    tree.visit_postorder(node_visit)



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tensorflow ops for updating target networks.

Tensorflow ops that are used to update a target network from a source network.
This is used in agents such as DQN or DPG, which use a target network that
changes more slowly than the online network, in order to improve stability.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports
import tensorflow as tf
from . import periodic_ops


def update_target_variables(
    target_variables,
    source_variables,
    tau=1.0,
    use_locking=False,
    name="update_target_variables",
):
    """Returns an op to update a list of target variables from source variables.

  The update rule is:
  `target_variable = (1 - tau) * target_variable + tau * source_variable`.

  Args:
    target_variables: a list of the variables to be updated.
    source_variables: a list of the variables used for the update.
    tau: weight used to gate the update. The permitted range is 0 < tau <= 1,
      with small tau representing an incremental update, and tau == 1
      representing a full update (that is, a straight copy).
    use_locking: use `tf.Variable.assign`'s locking option when assigning
      source variable values to target variables.
    name: sets the `name_scope` for this op.

  Raises:
    TypeError: when tau is not a Python float
    ValueError: when tau is out of range, or the source and target variables
      have different numbers or shapes.

  Returns:
    An op that executes all the variable updates.
  """
    if not isinstance(tau, float):
        raise TypeError("Tau has wrong type (should be float) {}".format(tau))
    if not 0.0 < tau <= 1.0:
        raise ValueError("Invalid parameter tau {}".format(tau))
    if len(target_variables) != len(source_variables):
        raise ValueError(
            "Number of target variables {} is not the same as "
            "number of source variables {}".format(
                len(target_variables), len(source_variables)
            )
        )

    same_shape = all(
        trg.get_shape() == src.get_shape()
        for trg, src in zip(target_variables, source_variables)
    )
    if not same_shape:
        raise ValueError(
            "Target variables don't have the same shape as source " "variables."
        )

    def update_op(target_variable, source_variable, tau):
        if tau == 1.0:
            return target_variable.assign(source_variable, use_locking)
        else:
            return target_variable.assign(
                tau * source_variable + (1.0 - tau) * target_variable, use_locking
            )

    with tf.compat.v1.name_scope(name, values=target_variables + source_variables):
        update_ops = [
            update_op(target_var, source_var, tau)
            for target_var, source_var in zip(target_variables, source_variables)
        ]
        return tf.group(name="update_all_variables", *update_ops)


def periodic_target_update(
    target_variables,
    source_variables,
    update_period,
    tau=1.0,
    use_locking=False,
    name="periodic_target_update",
):
    """Returns an op to periodically update a list of target variables.

  The `update_target_variables` op is executed every `update_period`
  executions of the `periodic_target_update` op.

  The update rule is:
  `target_variable = (1 - tau) * target_variable + tau * source_variable`.

  Args:
    target_variables: a list of the variables to be updated.
    source_variables: a list of the variables used for the update.
    update_period: inverse frequency with which to apply the update.
    tau: weight used to gate the update. The permitted range is 0 < tau <= 1,
      with small tau representing an incremental update, and tau == 1
      representing a full update (that is, a straight copy).
    use_locking: use `tf.variable.Assign`'s locking option when assigning
      source variable values to target variables.
    name: sets the `name_scope` for this op.

  Returns:
    An op that periodically updates `target_variables` with `source_variables`.
  """

    def update_op():
        return update_target_variables(
            target_variables, source_variables, tau, use_locking
        )

    with tf.compat.v1.name_scope(name, values=target_variables + source_variables):
        return periodic_ops.periodically(update_op, update_period)



import math
from typing import Any, List, Tuple

import numpy

from ..env import MathyEnv
from .policy_value_model import PolicyValueModel
from ..state import MathyEnvState, observations_to_window
from ..util import is_terminal_transition

EPS = 1e-8


class MCTS:
    """Monte-Carlo Tree Search is used for improving the actions selected
    for an agent by simulating a bunch of actions and returning the best
    based on the reward at the end of the simulated episodes."""

    env: MathyEnv
    # cpuct is a hyperparameter controlling the degree of exploration
    # (1.0 in Suragnair experiments.)
    cpuct: float
    num_mcts_sims: int
    # Set epsilon = 0 to disable dirichlet noise in root node.
    # e.g. for ExaminationRunner competitions
    epsilon: float
    dir_alpha: float

    def __init__(
        self,
        env: MathyEnv,
        model: PolicyValueModel,
        cpuct: float = 1.0,
        num_mcts_sims: int = 15,
        epsilon: float = 0.25,
        dir_alpha: float = 0.3,
    ):
        self.env = env
        self.model = model
        self.num_mcts_sims = num_mcts_sims
        self.cpuct = cpuct
        self.dir_alpha = dir_alpha
        self.epsilon = epsilon

        self.Qsa = {}  # stores Q values for s,a (as defined in the paper)
        self.Nsa = {}  # stores #times edge s,a was visited
        self.Ns = {}  # stores #times env_state s was visited
        self.Ps = {}  # stores initial policy (returned by neural net)
        self.Rs = {}  # stores observed rnn states at Ps

        self.Es = {}  # stores env.get_state_transition ended for env_state s
        self.Vs = {}  # stores env.get_valid_moves for env_state s

    def estimate_policy(
        self, env_state: MathyEnvState, rnn_state: List[Any], temp: float = 1.0
    ) -> Tuple[List[float], float]:
        """
        This function performs num_mcts_sims simulations of MCTS starting from
        env_state.

        Returns:
            probs: a policy vector where the probability of the ith action is
                   proportional to Nsa[(s,a)]**(1./temp)
            value: the mean value of this state across all rollouts
        """
        probs: List[float]
        values: List[float] = []
        for _ in range(self.num_mcts_sims):
            values.append(self.search(env_state, rnn_state, True))
        value = numpy.asarray(values).mean()

        s = self.env.to_hash_key(env_state)
        counts = []
        for a in range(self.env.get_agent_actions_count(env_state)):
            if (s, a) in self.Nsa:
                counts.append(self.Nsa[(s, a)])
            else:
                counts.append(0)

        if temp == 0:
            bestA = numpy.argmax(counts)
            probs = [0.0] * len(counts)
            probs[bestA] = 1.0
            return probs, value

        counts = [x ** (1.0 / temp) for x in counts]
        count_sum = float(sum(counts))
        if count_sum == 0.0:
            # Arg, no valid moves picked from the tree! Let's go ahead and make each
            valids = numpy.array(self.env.get_valid_moves(env_state))
            return list(valids / valids.sum()), value
        probs = [x / float(count_sum) for x in counts]
        return probs, value

    def search(self, env_state: MathyEnvState, rnn_state: List[Any], isRootNode=False):
        """
        This function performs one iteration of MCTS. It is recursively called
        until a leaf node is found. The action chosen at each node is one that
        has the maximum upper confidence bound as in the paper.

        Once a leaf node is found, the neural network is called to return an
        initial policy P and a value v for the state. This value is propogated
        up the search path. In case the leaf node is a terminal state, the
        outcome is propogated up the search path. The values of Ns, Nsa, Qsa are
        updated.

        Returns:
            v: the value of the current state
        """
        import tensorflow as tf

        s = self.env.to_hash_key(env_state)

        # if s not in self.Es:
        # print('calculating ending state for: {}'.format(s))
        self.Es[s] = self.env.get_state_transition(env_state, searching=True)
        if is_terminal_transition(self.Es[s]):
            # terminal node
            return self.Es[s].reward

        # This state does not have a predicted policy of value vector
        if s not in self.Ps or s not in self.Rs:
            # leaf node
            valids = self.env.get_valid_moves(env_state)
            obs = env_state.to_observation(
                valids,
                rnn_state_h=rnn_state[0],
                rnn_state_c=rnn_state[1],
                rnn_history_h=rnn_state[0],
            )
            observations = observations_to_window([obs]).to_inputs()
            out_policy, state_v = self.model.predict_next(observations)
            out_rnn_state = [
                tf.squeeze(self.model.embedding.state_h).numpy(),
                tf.squeeze(self.model.embedding.state_c).numpy(),
            ]
            self.Rs[s] = out_rnn_state
            self.Ps[s] = out_policy
            self.Vs[s] = valids
            self.Ns[s] = 0
            return state_v
        else:
            # Use the observed RNN state when this was cached
            out_rnn_state = self.Rs[s]

        valids = self.Vs[s]
        cur_best = -float("inf")
        all_best: List[int] = []
        # add Dirichlet noise for root node. set epsilon=0 for ExaminationRunner
        # competitions of trained models
        add_noise = isRootNode and self.epsilon > 0
        if add_noise:
            moves = self.env.get_valid_moves(env_state)
            noise = numpy.random.dirichlet([self.dir_alpha] * len(moves))

        # pick the action with the highest upper confidence bound
        i = -1
        for a in range(len(valids)):
            if valids[a]:
                i += 1
                if (s, a) in self.Qsa:
                    q = self.Qsa[(s, a)]
                    n_s_a = self.Nsa[(s, a)]
                else:
                    q = 0
                    n_s_a = 0

                p = self.Ps[s][a]
                if add_noise:
                    p = (1 - self.epsilon) * p + self.epsilon * noise[i]

                u = q + self.cpuct * p * math.sqrt(self.Ns[s]) / (1 + n_s_a)

                if u > cur_best:
                    cur_best = u
                    del all_best[:]
                    all_best.append(a)
                elif u == cur_best:
                    all_best.append(a)

        a = numpy.random.choice(all_best)
        next_s, _, _ = self.env.get_next_state(env_state, a, searching=True)
        state_v = self.search(next_s, out_rnn_state)

        # state key for next state
        state_key = (s, a)
        if state_key in self.Qsa:
            self.Qsa[state_key] = (
                self.Nsa[state_key] * self.Qsa[state_key] + state_v
            ) / (self.Nsa[state_key] + 1)
            self.Nsa[state_key] += 1

        else:
            self.Qsa[state_key] = state_v
            self.Nsa[state_key] = 1

        self.Ns[s] += 1
        return state_v



"""Mathy CLI
---

Command line application for interacting with Mathy agents and environments.
"""
from multiprocessing import cpu_count

import click
from wasabi import msg


@click.group()
@click.version_option()
def cli():
    """
    Mathy - https://mathy.ai

    Command line app for training and evaluating agents that transform
    expression trees using reinforcement learning.
    """


@cli.command("contribute")
def cli_contribute():
    """Learn about ways you can contribute to Mathy's development."""
    import webbrowser

    webbrowser.open("https://mathy.ai/contributing", new=2)


@cli.command("simplify")
@click.option(
    "model", "--model", default="mathy_alpha_sm", help="The path to a mathy model",
)
@click.option("agent", "--agent", default="a3c", help="one of 'a3c' or 'zero'")
@click.option(
    "thinking_steps",
    "--think",
    default=3,
    help="The number of steps to think about the problem before starting an episode",
)
@click.option(
    "max_steps",
    "--max-steps",
    default=20,
    help="The max number of steps before the episode is over",
)
@click.argument("problem", type=str)
def cli_simplify(
    agent: str, problem: str, model: str, thinking_steps: int, max_steps: int
):
    """Simplify an input polynomial expression."""
    setup_tf_env()

    from .models import load_model
    from .api import Mathy

    mt: Mathy = load_model(model)
    mt.simplify(problem=problem, max_steps=max_steps, thinking_steps=thinking_steps)


@cli.command("problems")
@click.argument("environment", type=str)
@click.option(
    "difficulty",
    "--difficulty",
    default="easy",
    help="One of 'easy', 'normal', or 'hard'",
)
@click.option("number", "--number", default=10, help="The number of problems to print")
def cli_print_problems(environment: str, difficulty: str, number: int):
    """Print a set of generated problems from a given environment.

    This is useful if you when developing new environment types for
    verifying that the problems you're generating take the form you
    expect. """
    import gym
    from mathy.envs.gym import MathyGymEnv

    env_name = f"mathy-{environment}-{difficulty}-v0"
    env: MathyGymEnv = gym.make(env_name)
    msg.divider(env_name)
    with msg.loading(f"Generating {number} problems..."):
        header = ("Complexity", "Is Valid", "Text")
        widths = (10, 8, 62)
        aligns = ("c", "c", "l")
        data = []
        for i in range(number):
            state, problem = env.mathy.get_initial_state(
                env.env_problem_args, print_problem=False
            )
            valid = False
            text = problem.text
            try:
                env.mathy.parser.parse(problem.text)
                valid = True
            except BaseException as error:
                text = f"parse failed for '{problem.text}' with error: {error}"
            data.append((problem.complexity, "✔" if valid else "✘", text,))
    msg.good(f"\nGenerated {number} problems!")

    print(msg.table(data, header=header, divider=True, widths=widths, aligns=aligns))


@cli.command("train")
@click.argument("agent")
@click.argument("topics")
@click.argument("folder")
@click.option(
    "transfer",
    "--transfer",
    default=None,
    help="Location of a model to initialize the agent's weights from",
)
@click.option(
    "difficulty",
    "--difficulty",
    default=None,
    help="the difficulty of problems to generate, 'easy','normal','hard'",
)
@click.option(
    "workers",
    "--workers",
    default=cpu_count(),
    type=int,
    help="Number of worker threads to use. More increases diversity of exp",
)
@click.option(
    "strategy",
    "--strategy",
    default="a3c",
    type=str,
    help="The action selection strategy to use",
)
@click.option(
    "units",
    "--units",
    default=512,
    type=int,
    help="Number of dimensions to use for math vectors and model dimensions",
)
@click.option(
    "lr",
    "--lr",
    default=3e-4,
    type=float,
    help="The learning rate to use with adam optimizer",
)
@click.option(
    "embeddings",
    "--embeddings",
    default=512,
    type=int,
    help="Number of dimensions to use for token embeddings",
)
@click.option(
    "use_lstm",
    "--use-lstm",
    type=bool,
    help="Whether to use the recurrent architecture or not",
)
@click.option(
    "rnn",
    "--rnn",
    default=128,
    type=int,
    help="Number of dimensions to use for RNN state",
)
@click.option(
    "episodes",
    "--episodes",
    default=None,
    type=int,
    help="Maximum number of episodes to run",
)
@click.option(
    "mcts_sims",
    "--mcts-sims",
    default=10,
    type=int,
    help="Number of rollouts per timestep when using MCTS",
)
@click.option(
    "show",
    "--show",
    default=False,
    is_flag=True,
    help="Show the agents step-by-step directions",
)
@click.option(
    "profile",
    "--profile",
    default=False,
    is_flag=True,
    help="Set to gather profiler outputs for workers",
)
@click.option(
    "self_play_problems",
    "--self-play-problems",
    default=100,
    help="The number of self-play problems per gather/training iteration",
)
@click.option(
    "training_iterations",
    "--training-iterations",
    default=10,
    help="The max number of time to perform gather/training loops for the zero agent",
)
@click.option(
    "verbose",
    "--verbose",
    default=False,
    is_flag=True,
    help="Display verbose log items",
)
def cli_train(
    agent: str,
    topics: str,
    folder: str,
    transfer: str,
    difficulty: str,
    strategy: str,
    workers: int,
    units: int,
    embeddings: int,
    rnn: int,
    profile: bool,
    episodes: int,
    mcts_sims: int,
    lr: float,
    show: bool,
    verbose: bool,
    training_iterations: int,
    self_play_problems: int,
    use_lstm: bool,
):
    """Train an agent to solve math problems and save the model.

    Arguments:

    "agent" is one of the known agent types, either "a3c" or "zero".

    "topics" is a comma separated list of topic names to work problems from.

    "folder" is the output location to store the model and tensorboard logs.
    e.g. "/tmp/training/custom_agent/"

    """
    topics_list = topics.split(",")

    if agent == "a3c":
        setup_tf_env()

        from .agents.a3c import A3CAgent, A3CConfig

        args = A3CConfig(
            verbose=verbose,
            difficulty=difficulty,
            action_strategy=strategy,
            topics=topics_list,
            lstm_units=rnn,
            units=units,
            embedding_units=embeddings,
            lr=lr,
            mcts_sims=mcts_sims,
            model_dir=folder,
            init_model_from=transfer,
            num_workers=workers,
            profile=profile,
            print_training=show,
        )
        if episodes is not None:
            args.max_eps = episodes
        if use_lstm is not None:
            args.use_lstm = use_lstm
        instance = A3CAgent(args)
        instance.train()
    elif agent == "zero":
        setup_tf_env(use_mp=workers > 1)
        from .agents.zero import SelfPlayConfig, self_play_runner

        self_play_cfg = SelfPlayConfig(
            verbose=verbose,
            difficulty=difficulty,
            topics=topics_list,
            lr=lr,
            lstm_units=rnn,
            units=units,
            embedding_units=embeddings,
            mcts_sims=mcts_sims,
            model_dir=folder,
            init_model_from=transfer,
            num_workers=workers,
            training_iterations=training_iterations,
            self_play_problems=self_play_problems,
            print_training=show,
            profile=profile,
        )
        if use_lstm is not None:
            self_play_cfg.use_lstm = use_lstm
        if episodes is not None:
            self_play_cfg.max_eps = episodes

        self_play_runner(self_play_cfg)


def setup_tf_env(use_mp=False):
    if use_mp:
        setup_tf_env_mp()
    import os

    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "5"
    import random
    import numpy as np
    import tensorflow as tf

    random.seed(1337)
    np.random.seed(1337)
    tf.random.set_seed(1337)

    tf.compat.v1.logging.set_verbosity("CRITICAL")


def setup_tf_env_mp():
    """Create a sub-process and import Tensorflow inside of it
    so that the library is loaded first in a subprocess. This is
    the hacky way we get multiprocessing to work reliably. |o_O|"""
    import multiprocessing

    def worker():
        import tensorflow as tf

        return 0

    proc = multiprocessing.Process(target=worker)
    proc.start()
    proc.join()


if __name__ == "__main__":
    cli()



from colr import color
import datetime


def truncate(value):
    return float("%.3f" % (float(value)))


def record(
    episode,
    episode_reward,
    worker_idx,
    global_ep_reward,
    result_queue,
    pi_loss,
    value_loss,
    entropy_loss,
    aux_losses,
    total_loss,
    num_steps,
    env_name,
):
    """Helper function to store score and print statistics.
  Arguments:
    episode: Current episode
    episode_reward: Reward accumulated over the current episode
    worker_idx: Which thread (worker)
    global_ep_reward: The moving average of the global reward
    result_queue: Queue storing the moving average of the scores
    total_loss: The total loss accumualted over the current episode
    num_steps: The number of steps the episode took to complete
  """

    now = datetime.datetime.now().strftime("%H:%M:%S")

    global_ep_reward = global_ep_reward * 0.99 + episode_reward * 0.01

    fore = "green" if episode_reward > 0.0 else "red"

    losses = [
        "total: {:<8}".format(truncate(total_loss)),
        "pi: {:<8}".format(truncate(pi_loss)),
        "v: {:<8}".format(truncate(value_loss)),
        "h: {:<8}".format(truncate(entropy_loss)),
    ]
    if isinstance(aux_losses, dict):
        for k in aux_losses.keys():
            losses.append("{:<4}: {:<8}".format(k, truncate(aux_losses[k])))

    heading = "{:<8} {:<3} {:<8} {:<10}".format(
        now, f"w{worker_idx}", f"ep: {episode}", f"steps: {num_steps}"
    )
    rewards = "r_avg: {:<6} r_ep: {:<6}".format(
        truncate(global_ep_reward), truncate(episode_reward)
    )
    loss_str = " ".join(losses)
    print(
        color(f"{heading} {rewards} {loss_str} [{env_name}]", fore=fore, style="bright")
    )
    result_queue.put(global_ep_reward)
    return global_ep_reward



from gym.envs.registration import register

from ..binomial_distribute import BinomialDistribute
from ...types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register


class GymBinomialDistribution(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs):
        super(GymBinomialDistribution, self).__init__(
            env_class=BinomialDistribute,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class BinomialsEasy(GymBinomialDistribution):
    def __init__(self, **kwargs):
        super(BinomialsEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class BinomialsNormal(GymBinomialDistribution):
    def __init__(self, **kwargs):
        super(BinomialsNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class BinomialsHard(GymBinomialDistribution):
    def __init__(self, **kwargs):
        super(BinomialsHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(id="mathy-binomial-easy-v0", entry_point="mathy.envs.gym:BinomialsEasy")
safe_register(id="mathy-binomial-normal-v0", entry_point="mathy.envs.gym:BinomialsNormal")
safe_register(id="mathy-binomial-hard-v0", entry_point="mathy.envs.gym:BinomialsHard")



from typing import Optional, Tuple

from ..core.expressions import (
    AddExpression,
    BinaryExpression,
    ConstantExpression,
    MultiplyExpression,
    VariableExpression,
)
from ..core.rule import BaseRule


class ConstantsSimplifyRule(BaseRule):
    """Given a binary operation on two constants, simplify to the resulting
    constant expression"""

    @property
    def name(self):
        return "Constant Arithmetic"

    @property
    def code(self):
        return "CA"

    POS_SIMPLE: str = "simple"
    POS_SIMPLE_VAR_MULT: str = "simple_var_multiply"
    POS_CHAINED_RIGHT: str = "chained_right"
    POS_CHAINED_RIGHT_LEFT: str = "chained_right_left"
    POS_CHAINED_RIGHT_LEFT_LEFT: str = "chained_right_left_left"
    POS_CHAINED_LEFT_LEFT_RIGHT: str = "chained_left_left_right"
    POS_CHAINED_RIGHT_DEEP: str = "chained_right_deep"

    def get_type(
        self, node
    ) -> Optional[Tuple[str, ConstantExpression, ConstantExpression]]:
        """Determine the configuration of the tree for this transformation.

        Support the three types of tree configurations:
         - Simple is where the node's left and right children are exactly
           constants linked by an add operation.
         - Chained Right is where the node's left child is a constant, but the right
           child is another binary operation of the same type. In this case the left
           child of the next binary node is the target.

        Structure:
         - Simple
            * node(add),node.left(const),node.right(const)
         - Chained Right
            * node(add),node.left(const),node.right(add),node.right.left(const)
         - Chained Right Deep
            * node(add),node.left(const),node.right(add),node.right.left(const)
        """
        # Check simple case of left/right child binary op with constants
        # (4 * 2) + 3
        if (
            isinstance(node, BinaryExpression)
            and isinstance(node.left, ConstantExpression)
            and isinstance(node.right, ConstantExpression)
        ):
            return ConstantsSimplifyRule.POS_SIMPLE, node.left, node.right

        # Check for const * var * const
        # (4n * 2) + 3
        if (
            isinstance(node, MultiplyExpression)
            and isinstance(node.left, MultiplyExpression)
            and isinstance(node.left.left, ConstantExpression)
            and isinstance(node.left.right, VariableExpression)
            and isinstance(node.right, ConstantExpression)
        ):
            return ConstantsSimplifyRule.POS_SIMPLE_VAR_MULT, node.left.left, node.right

        # Check for a continuation to the right that's more than one node
        # e.g. "5 * (8h * t)" = "40h * t"
        if (
            isinstance(node, BinaryExpression)
            and isinstance(node.left, ConstantExpression)
            and isinstance(node.right, BinaryExpression)
            and isinstance(node.right.left, BinaryExpression)
            and isinstance(node.right.left.left, ConstantExpression)
        ):
            # Add/Multiply continuations are okay
            if (
                isinstance(node, AddExpression)
                and isinstance(node.right, AddExpression)
                and isinstance(node.right.left, AddExpression)
                or isinstance(node, MultiplyExpression)
                and isinstance(node.right, MultiplyExpression)
                and isinstance(node.right.left, MultiplyExpression)
            ):
                return (
                    ConstantsSimplifyRule.POS_CHAINED_RIGHT_DEEP,
                    node.left,
                    node.right.left.left,
                )

        # Check for a continuation to the right
        # "(7 * 10y^3) * x"
        if (
            isinstance(node, BinaryExpression)
            and isinstance(node.left, ConstantExpression)
            and isinstance(node.right, BinaryExpression)
            and isinstance(node.right.left, ConstantExpression)
        ):
            # Add/Multiply continuations are okay
            if (
                isinstance(node, AddExpression)
                and isinstance(node.right, AddExpression)
                or isinstance(node, MultiplyExpression)
                and isinstance(node.right, MultiplyExpression)
            ):
                return (
                    ConstantsSimplifyRule.POS_CHAINED_RIGHT,
                    node.left,
                    node.right.left,
                )

        # Check for a continuation to the right
        # "(7q * 10y^3) * x"
        if (
            isinstance(node, MultiplyExpression)
            and isinstance(node.left, MultiplyExpression)
            and isinstance(node.left.left, ConstantExpression)
            and isinstance(node.right, MultiplyExpression)
            and isinstance(node.right.left, ConstantExpression)
        ):
            return (
                ConstantsSimplifyRule.POS_CHAINED_RIGHT_LEFT,
                node.left.left,
                node.right.left,
            )

        # Check for variable terms with constants on the left and right
        # "792z^4 * 490f * q^3"
        #   ^--------^
        if (
            isinstance(node, MultiplyExpression)
            and isinstance(node.left, MultiplyExpression)
            and isinstance(node.left.left, ConstantExpression)
            and isinstance(node.right, MultiplyExpression)
            and isinstance(node.right.left, MultiplyExpression)
            and isinstance(node.right.left.left, ConstantExpression)
        ):
            return (
                ConstantsSimplifyRule.POS_CHAINED_RIGHT_LEFT_LEFT,
                node.left.left,
                node.right.left.left,
            )

        # Check for variable terms with constants nested on the left and right
        # "(u^3 * 36c^6) * 7u^3"
        #         ^--------^
        if (
            isinstance(node, MultiplyExpression)
            and isinstance(node.left, MultiplyExpression)
            and isinstance(node.left.right, MultiplyExpression)
            and isinstance(node.left.right.left, ConstantExpression)
            and isinstance(node.right, MultiplyExpression)
            and isinstance(node.right.left, ConstantExpression)
        ):
            return (
                ConstantsSimplifyRule.POS_CHAINED_LEFT_LEFT_RIGHT,
                node.left.right.left,
                node.right.left,
            )

        return None

    def can_apply_to(self, node):
        return self.get_type(node) is not None

    def apply_to(self, node):
        change = super().apply_to(node)
        arrangement, left_const, right_const = self.get_type(node)
        change.save_parent()
        if arrangement == ConstantsSimplifyRule.POS_SIMPLE:
            result = ConstantExpression(node.evaluate())
        if arrangement == ConstantsSimplifyRule.POS_SIMPLE_VAR_MULT:
            assert isinstance(node, MultiplyExpression)
            assert isinstance(node.left.right, VariableExpression)
            value = ConstantExpression(
                MultiplyExpression(left_const, right_const).evaluate()
            )
            result = MultiplyExpression(value, node.left.right)
        elif arrangement == ConstantsSimplifyRule.POS_CHAINED_LEFT_LEFT_RIGHT:
            assert isinstance(node, MultiplyExpression)
            value = ConstantExpression(
                MultiplyExpression(left_const, right_const).evaluate()
            )
            result = MultiplyExpression(
                node.left.left,
                MultiplyExpression(
                    MultiplyExpression(value, node.left.right.right), node.right.right
                ),
            )

        elif arrangement == ConstantsSimplifyRule.POS_CHAINED_RIGHT_LEFT:
            value = ConstantExpression(
                MultiplyExpression(left_const, right_const).evaluate()
            )
            value = MultiplyExpression(value, node.left.right)
            result = MultiplyExpression(value, node.right.right)
        elif arrangement == ConstantsSimplifyRule.POS_CHAINED_RIGHT_LEFT_LEFT:
            value = ConstantExpression(
                MultiplyExpression(left_const, right_const).evaluate()
            )
            value = MultiplyExpression(value, node.left.right)
            result = MultiplyExpression(
                value, MultiplyExpression(node.right.left.right, node.right.right)
            )
        elif arrangement == ConstantsSimplifyRule.POS_CHAINED_RIGHT:
            if isinstance(node, AddExpression):
                value = ConstantExpression(
                    AddExpression(left_const, right_const).evaluate()
                )
                result = AddExpression(value, node.right.right)
            elif isinstance(node, MultiplyExpression):
                value = ConstantExpression(
                    MultiplyExpression(left_const, right_const).evaluate()
                )
                result = MultiplyExpression(value, node.right.right)
            else:
                raise NotImplementedError(
                    f"can't deal with operand of {type(node)} type"
                )
        elif arrangement == ConstantsSimplifyRule.POS_CHAINED_RIGHT_DEEP:
            if isinstance(node, AddExpression):
                value = ConstantExpression(
                    AddExpression(left_const, right_const).evaluate()
                )
                result = AddExpression(
                    AddExpression(value, node.right.left.right), node.right.right
                )
            elif isinstance(node, MultiplyExpression):
                value = ConstantExpression(
                    MultiplyExpression(left_const, right_const).evaluate()
                )
                result = MultiplyExpression(
                    MultiplyExpression(value, node.right.left.right), node.right.right
                )
            else:
                raise NotImplementedError(
                    f"can't deal with operand of {type(node)} type"
                )
        result.set_changed()
        return change.done(result)



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TensorFlow ops for various distribution projection operations.

All ops support multidimensional tensors. All dimensions except for the last
one can be considered as batch dimensions. They are processed in parallel
and are fully independent. The last dimension represents the number of bins.

The op supports broadcasting across all dimensions except for the last one.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_probability as tfp


def l2_project(support, weights, new_support):
    """Projects distribution (support, weights) onto new_support.

  Args:
    support: Tensor defining support of a categorical distribution(s). Must be
      of rank 1 or of the same rank as `weights`. The size of the last dimension
      has to match that of `weights`.
    weights: Tensor defining weights on the support points.
    new_support: Tensor holding positions of a new support.
  Returns:
    Projection of (support, weights) onto the new_support.
  """
    return gen_distribution_ops.project_distribution(support, weights, new_support, 1)


def hard_cumulative_project(support, weights, new_support, reverse):
    """Produces a cumulative categorical distribution on a new support.

  Args:
    support: Tensor defining support of a categorical distribution(s). Must be
      of rank 1 or of the same rank as `weights`. The size of the last dimension
      has to match that of `weights`.
    weights: Tensor defining weights on the support points.
    new_support: Tensor holding positions of a new support.
    reverse: Whether to evalute cumulative from the left or right.
  Returns:
    Cumulative distribution on the supplied support.
    The foolowing invariant is maintained across the last dimension:
    result[i] = (sum_j weights[j] for all j where support[j] < new_support[i])
                if reverse == False else
                (sum_j weights[j] for all j where support[j] > new_support[i])
  """
    return gen_distribution_ops.project_distribution(
        support, weights, new_support, 3 if reverse else 2
    )


def factorised_kl_gaussian(
    dist1_mean,
    dist1_covariance_or_scale,
    dist2_mean,
    dist2_covariance_or_scale,
    both_diagonal=False,
):
    """Compute the KL divergence KL(dist1, dist2) between two Gaussians.

  The KL is factorised into two terms - `kl_mean` and `kl_cov`. This
  factorisation is specific to multivariate gaussian distributions and arises
  from its analytic form.
  Specifically, if we assume two multivariate Gaussian distributions with rank
  k and means, M1 and M2 and variance S1 and S2, the analytic KL can be written
  out as:

  D_KL(N0 || N1) = 0.5 * (tr(inv(S1) * S0) + ln(det(S1)/det(S0)) - k +
                         (M1 - M0).T * inv(S1) * (M1 - M0))

  The terms on the first row correspond to the covariance factor and the terms
  on the second row correspond to the mean factor in the factorized KL.
  These terms can thus be used to independently control how much the mean and
  covariance between the two gaussians can vary.

  This implementation ensures that gradient flow is equivalent to calling
  `tfp.distributions.kl_divergence` once.

  More details on the equation can be found here:
  https://stats.stackexchange.com/questions/60680/kl-divergence-between-two-multivariate-gaussians


  Args:
    dist1_mean: The mean of the first Multivariate Gaussian distribution.
    dist1_covariance_or_scale: The covariance or scale of the first Multivariate
      Gaussian distribution. In cases where *both* distributions are Gaussians
      with diagonal covariance matrices (for instance, if both are instances of
      `tfp.distributions.MultivariateNormalDiag`), then the `scale` can be
      passed in instead and the `both_diagonal` flag must be set to `True`.
      A more efficient sparse computation path is used in this case. For all
      other cases, the full covariance matrix must be passed in.
    dist2_mean: The mean of the second Multivariate Gaussian distribution.
    dist2_covariance_or_scale: The covariance or scale tensor of the second
      Multivariate Gaussian distribution, as for `dist1_covariance_or_scale`.
    both_diagonal: A `bool` indicating that both dist1 and dist2 are diagonal
      matrices. A more efficient sparse computation is used in this case.

  Returns:
    A tuple consisting of (`kl_mean`, `kl_cov`) which correspond to the mean and
    the covariance factorisation of the KL.
  """
    if both_diagonal:
        dist1_mean_rank = dist1_mean.get_shape().ndims
        dist1_covariance_or_scale.get_shape().assert_has_rank(dist1_mean_rank)
        dist2_mean_rank = dist2_mean.get_shape().ndims
        dist2_covariance_or_scale.get_shape().assert_has_rank(dist2_mean_rank)

        dist_type = tfp.distributions.MultivariateNormalDiag
    else:
        dist_type = tfp.distributions.MultivariateNormalFullCovariance

    # Recreate the distributions but with stop gradients on the mean and cov.
    dist1_stop_grad_mean = dist_type(
        tf.stop_gradient(dist1_mean), dist1_covariance_or_scale
    )
    dist2 = dist_type(dist2_mean, dist2_covariance_or_scale)

    # Now create a third distribution with the mean of dist1 and the variance of
    # dist2 and appropriate stop_gradients.
    dist3 = dist_type(dist1_mean, dist2_covariance_or_scale)
    dist3_stop_grad_mean = dist_type(
        tf.stop_gradient(dist1_mean), dist2_covariance_or_scale
    )

    # Finally get the two components of the KL between dist1 and dist2
    # using dist3
    kl_mean = tfp.distributions.kl_divergence(dist3, dist2)
    kl_cov = tfp.distributions.kl_divergence(dist1_stop_grad_mean, dist3_stop_grad_mean)
    return kl_mean, kl_cov



from typing import Optional

from ..envs import PolySimplify
from ..state import MathyWindowObservation, observations_to_window
from .base_config import BaseConfig


def example() -> MathyWindowObservation:
    """Helper to return a random Window observation that can be 
    passed forward through a Mathy model. """
    env = PolySimplify()
    state = env.get_initial_state()[0]
    observation = env.state_to_observation(state, rnn_size=BaseConfig().lstm_units)
    return observations_to_window([observation])



from ...agents.base_config import BaseConfig


class A3CConfig(BaseConfig):
    # Update frequencey for the Worker to sync with the Main model.
    #
    # Indicates the maximum number of steps to take in an episode before
    # syncing the replay buffer and gradients.
    update_gradients_every: int = 6

    normalization_style: str = "layer"

    # How many times to think about the initial state before acting.
    # (intuition) is that the LSTM updates the state each time it processes
    # the init sequence meaning that it gets more time to fine-tune the hidden
    # and cell states for the particular problem.
    num_thinking_steps_begin: int = 3

    # Strategy for introducing MCTS into the A3C agent training process
    #
    #   - "a3c" Do not use MCTS when training the A3C agent
    #   - "mcts_worker_0" uses MCTS for observations on the greediest worker. This
    #                usually looks the best visually, because mcts_worker_0 is print
    #                to stdout, so it results in lots of green. It's unclear that
    #                it helps improve the A3C agent performance after MCTS is removed.
    #   - "mcts_worker_n" uses MCTS for observations on all workers except worker_0.
    #                This adds the strength of MCTS to observation gathering, without
    #                biasing the observed strength of the model (because only worker_0)
    #                reports statistics.
    action_strategy = "a3c"
    # MCTS provides higher quality observations at extra computational cost.
    mcts_sims: int = 200

    main_worker_use_epsilon = False
    e_greedy_min = 0.01
    e_greedy_max = 0.3
    # Worker's sleep this long between steps to allow
    # other threads time to process. This is useful for
    # running more threads than you have processors to
    # get a better diversity of experience.
    worker_wait: float = 0.01

    # The number of worker agents to create.
    num_workers: int = 3

    # NOTE: scaling down h_loss is observed to be important to keep it from
    #       destabilizing the overall loss when it grows very small
    entropy_loss_scaling = 0.05
    # Whether to scale entropy loss so it's 0-1
    normalize_entropy_loss = True

    # How much to scale down loss values from auxiliary tasks
    aux_tasks_weight_scale = 0.1
    # The lambda value for generalized lambda returns to calculate value loss
    # 0.0 = bootstrap values, 1.0 = discounted
    td_lambda: float = 0.2

    # The "Teacher" will start evaluating after this many initial episodes
    teacher_start_evaluations_at_episode = 50
    # The "Teacher" evaluates the win/loss record of the agent every (n) episodes
    teacher_evaluation_steps = 5
    # If the agent wins >= this value, promote to the next difficulty class
    # Wild-ass guess inspired by:
    # https://uanews.arizona.edu/story/learning-optimized-when-we-fail-15-time
    # If 85 is optimal, when you go beyond 85 + buffer it's time to move up... |x_X|
    teacher_promote_wins = 0.95
    # If the agent loses >= this value, demot to the previous difficulty class
    teacher_demote_wins = 0.60

    # When profile is true, each A3C worker thread will output a .profile
    # file in the model save path when it exits.
    profile: bool = False

    # Verbose setting to print out worker_0 training steps. Useful for trying
    # to find problems.
    print_training: bool = False

    # Print mode for output. "terminal" is the default, also supports "attention"
    # NOTE: attention is gone (like... the layer)
    print_mode: str = "terminal"



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Periodic execution ops.

It is very common in Reinforcement Learning for certain ops to only need to be
executed periodically, for example: once every N agent steps. The ops below
support this common use-case by wrapping a subgraph as a periodic op that only
actually executes the underlying computation once every N evaluations of the op,
behaving as a no-op in all other calls.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports
import tensorflow as tf


def periodically(body, period, name="periodically"):
    """Periodically performs a tensorflow op.

  The body tensorflow op will be executed every `period` times the periodically
  op is executed. More specifically, with `n` the number of times the op has
  been executed, the body will be executed when `n` is a non zero positive
  multiple of `period` (i.e. there exist an integer `k > 0` such that
  `k * period == n`).

  If `period` is 0 or `None`, it would not perform any op and would return a
  `tf.no_op()`.

  Args:
    body: callable that returns the tensorflow op to be performed every time
      an internal counter is divisible by the period. The op must have no
      output (for example, a tf.group()).
    period: inverse frequency with which to perform the op.
    name: name of the variable_scope.

  Raises:
    TypeError: if body is not a callable.
    ValueError: if period is negative.

  Returns:
    An op that periodically performs the specified op.
  """
    if not callable(body):
        raise TypeError("body must be callable.")

    if period is None or period == 0:
        return tf.no_op()

    if period < 0:
        raise ValueError("period cannot be less than 0.")

    if period == 1:
        return body()

    with tf.compat.v1.variable_scope(None, default_name=name):
        counter = tf.compat.v1.get_variable(
            "counter",
            shape=[],
            dtype=tf.int64,
            trainable=False,
            initializer=tf.compat.v1.constant_initializer(period, dtype=tf.int64),
        )

        def _wrapped_body():
            with tf.control_dependencies([body()]):
                # Done the deed, resets the counter.
                return counter.assign(1)

        update = tf.cond(
            pred=tf.equal(counter, period),
            true_fn=_wrapped_body,
            false_fn=lambda: counter.assign_add(1),
        )

    return update



from pydantic import BaseModel
from typing import NamedTuple, Any
from ...state import MathyObservation


class EpisodeSummary(BaseModel):
    text: str
    complexity: int
    duration: float
    reward: float
    solved: bool


class EpisodeHistory(NamedTuple):
    """A tuple of items related to an episode timestep."""

    text: str
    action: int
    reward: float
    discounted: float
    terminal: bool
    observation: MathyObservation
    pi: Any
    value: float


# fmt: off
EpisodeHistory.text.__doc__ = "The problem text at this state" # noqa
EpisodeHistory.action.__doc__ = "The action taken" # noqa
EpisodeHistory.reward.__doc__ = "The undiscounted reward observed" # noqa
EpisodeHistory.discounted.__doc__ = "The final discounted reward observed" # noqa
EpisodeHistory.terminal.__doc__ = "0/1 whether this state was terminal" # noqa
EpisodeHistory.observation.__doc__ = "MathyObservation for training" # noqa
EpisodeHistory.pi.__doc__ = "weighted distribution over the actions" # noqa
EpisodeHistory.value.__doc__ = "predicted value for the current state" # noqa
# fmt: on



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Ops to implement gradient clipping."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def huber_loss(input_tensor, quadratic_linear_boundary, name=None):
    """Calculates huber loss of `input_tensor`.

  For each value x in `input_tensor`, the following is calculated:

  ```
    0.5 * x^2                  if |x| <= d
    0.5 * d^2 + d * (|x| - d)  if |x| > d
  ```

  where d is `quadratic_linear_boundary`.

  When `input_tensor` is a loss this results in a form of gradient clipping.
  This is, for instance, how gradients are clipped in DQN and its variants.

  Args:
    input_tensor: `Tensor`, input values to calculate the huber loss on.
    quadratic_linear_boundary: `float`, the point where the huber loss function
      changes from a quadratic to linear.
    name: `string`, name for the operation (optional).

  Returns:
    `Tensor` of the same shape as `input_tensor`, containing values calculated
    in the manner described above.

  Raises:
    ValueError: if quadratic_linear_boundary <= 0.
  """
    if quadratic_linear_boundary < 0:
        raise ValueError("quadratic_linear_boundary must be > 0.")

    with tf.compat.v1.name_scope(
        name,
        default_name="huber_loss",
        values=[input_tensor, quadratic_linear_boundary],
    ):
        abs_x = tf.abs(input_tensor)
        delta = quadratic_linear_boundary
        quad = tf.minimum(abs_x, delta)
        # The following expression is the same in value as
        # tf.maximum(abs_x - delta, 0), but importantly the gradient for the
        # expression when abs_x == delta is 0 (for tf.maximum it would be 1). This
        # is necessary to avoid doubling the gradient, since there is already a
        # non-zero contribution to the gradient from the quadratic term.
        lin = abs_x - quad
        return 0.5 * quad ** 2 + delta * lin



from enum import IntEnum
from typing import Any, Dict, List, NamedTuple, Optional

import numpy as np

from .core.expressions import ConstantExpression, MathExpression, MathTypeKeys
from .core.parser import ExpressionParser
from .util import pad_array

PROBLEM_TYPE_HASH_BUCKETS = 128

NodeIntList = List[int]
NodeValuesFloatList = List[float]
NodeMaskIntList = List[int]
ProblemTypeIntList = List[int]
ProblemTimeFloatList = List[float]
RNNStateFloatList = List[float]


WindowNodeIntList = List[NodeIntList]
WindowNodeMaskIntList = List[NodeMaskIntList]
WindowNodeValuesFloatList = List[NodeValuesFloatList]
WindowProblemTypeIntList = List[ProblemTypeIntList]
WindowProblemTimeFloatList = List[ProblemTimeFloatList]
WindowRNNStateFloatList = List[RNNStateFloatList]


# Input type for mathy models
MathyInputsType = List[Any]


class ObservationFeatureIndices(IntEnum):
    nodes = 0
    mask = 1
    values = 2
    type = 3
    time = 4
    rnn_state_h = 5
    rnn_state_c = 6
    rnn_history_h = 7


class MathyObservation(NamedTuple):
    """A featurized observation from an environment state."""

    nodes: NodeIntList
    mask: NodeMaskIntList
    values: NodeValuesFloatList
    type: ProblemTypeIntList
    time: ProblemTimeFloatList
    rnn_state_h: RNNStateFloatList
    rnn_state_c: RNNStateFloatList
    rnn_history_h: RNNStateFloatList


# fmt: off
MathyObservation.nodes.__doc__ = "tree node types in the current environment state shape=[n,]" # noqa
MathyObservation.mask.__doc__ = "0/1 mask where 0 indicates an invalid action shape=[n,]" # noqa
MathyObservation.values.__doc__ = "tree node value sequences, with non number indices set to 0.0 shape=[n,]" # noqa
MathyObservation.type.__doc__ = "two column hash of problem environment type shape=[2,]" # noqa
MathyObservation.time.__doc__ = "float value between 0.0-1.0 for how much time has passed shape=[1,]" # noqa
MathyObservation.rnn_state_h.__doc__ = "rnn state_h shape=[dimensions]" # noqa
MathyObservation.rnn_state_c.__doc__ = "rnn state_c shape=[dimensions]" # noqa
MathyObservation.rnn_history_h.__doc__ = "rnn historical state_h shape=[dimensions]" # noqa
# fmt: on


class MathyWindowObservation(NamedTuple):
    """A featurized observation from an n-step sequence of environment states."""

    nodes: WindowNodeIntList
    mask: WindowNodeMaskIntList
    values: WindowNodeValuesFloatList
    type: WindowProblemTypeIntList
    time: WindowProblemTimeFloatList
    rnn_state_h: WindowRNNStateFloatList
    rnn_state_c: WindowRNNStateFloatList
    rnn_history_h: WindowRNNStateFloatList

    def to_inputs(self, as_tf_tensor: bool = True) -> MathyInputsType:
        import tensorflow as tf

        def to_res(in_value):
            if as_tf_tensor is True:
                return tf.convert_to_tensor(in_value, dtype=tf.float32)
            return np.asarray(in_value, dtype="float32")

        result = [
            to_res(self.nodes),
            to_res(self.mask),
            to_res(self.values),
            to_res(self.type),
            to_res(self.time),
            to_res(self.rnn_state_h),
            to_res(self.rnn_state_c),
            to_res(self.rnn_history_h),
        ]
        for r in result:
            for s in r.shape:
                assert s is not None
        return result

    def to_input_shapes(self) -> List[Any]:
        import tensorflow as tf

        result = [
            np.asarray(self.nodes).shape,
            np.asarray(self.mask).shape,
            np.asarray(self.values).shape,
            np.asarray(self.type).shape,
            np.asarray(self.time).shape,
            np.asarray(self.rnn_state_h).shape,
            np.asarray(self.rnn_state_c).shape,
            np.asarray(self.rnn_history_h).shape,
        ]
        return result


# fmt: off
MathyWindowObservation.nodes.__doc__ = "n-step list of node sequences `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.mask.__doc__ = "n-step list of node sequence masks `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.values.__doc__ = "n-step list of value sequences, with non number indices set to 0.0 `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.type.__doc__ = "n-step problem type hash `shape=[n, 2]`" # noqa
MathyWindowObservation.time.__doc__ = "float value between 0.0-1.0 for how much time has passed `shape=[n, 1]`" # noqa
MathyWindowObservation.rnn_state_h.__doc__ = "n-step rnn state `shape=[n, dimensions]`" # noqa
MathyWindowObservation.rnn_state_c.__doc__ = "n-step rnn state `shape=[n, dimensions]`" # noqa
MathyWindowObservation.rnn_history_h.__doc__ = "n-step rnn historical state `shape=[n, dimensions]`" # noqa
# fmt: on


class MathyEnvStateStep(NamedTuple):
    """Capture summarized environment state for a previous timestep so the
    agent can use context from its history when making new predictions."""

    raw: str
    focus: int
    action: int


# fmt: off
MathyEnvStateStep.raw.__doc__ = "the input text at the timestep" # noqa
MathyEnvStateStep.focus.__doc__ = "the index of the node that is acted on" # noqa
MathyEnvStateStep.action.__doc__ = "the action taken" # noqa
# fmt: on


_cached_placeholder: Optional[RNNStateFloatList] = None
_problem_hash_cache: Optional[Dict[str, List[int]]] = None


def rnn_placeholder_state(rnn_size: int) -> RNNStateFloatList:
    """Create a placeholder state for the RNN hidden states in an observation. This
    is useful because we don't always know the RNN state when we initialize an
    observation. """
    global _cached_placeholder
    # Not in cache, or rnn_size differs
    if _cached_placeholder is None or len(_cached_placeholder) != rnn_size:
        _cached_placeholder = [0.0] * rnn_size
    return _cached_placeholder


def observations_to_window(
    observations: List[MathyObservation],
) -> MathyWindowObservation:
    """Combine a sequence of observations into an observation window"""
    output = MathyWindowObservation(
        nodes=[],
        mask=[],
        values=[],
        type=[],
        time=[],
        rnn_state_h=[],
        rnn_state_c=[],
        rnn_history_h=[],
    )
    sequence_lengths: List[int] = []
    max_mask_length: int = 0
    for i in range(len(observations)):
        obs = observations[i]
        sequence_lengths.append(len(obs.nodes))
        sequence_lengths.append(len(obs.values))
        max_mask_length = max(max_mask_length, len(obs.mask))
    max_length: int = max(sequence_lengths)

    for obs in observations:
        output.nodes.append(pad_array(obs.nodes, max_length, MathTypeKeys["empty"]))
        output.mask.append(pad_array(obs.mask, max_mask_length, 0))
        output.values.append(pad_array(obs.values, max_length, 0.0))
        output.type.append(pad_array([], max_length, obs.type))
        output.time.append(pad_array([], max_length, obs.time))
        output.rnn_state_h.append(obs.rnn_state_h)
        output.rnn_state_c.append(obs.rnn_state_c)
        output.rnn_history_h.append(obs.rnn_history_h)
    return output


class MathyEnvState(object):
    """Class for holding environment state and extracting features
    to be passed to the policy/value neural network.

    Mutating operations all return a copy of the environment adapter
    with its own state.

    This allocation strategy requires more memory but removes a class
    of potential issues around unintentional sharing of data and mutation
    by two different sources.
    """

    agent: "MathyAgentState"
    parser: ExpressionParser
    max_moves: int
    num_rules: int

    def __init__(
        self,
        state: Optional["MathyEnvState"] = None,
        problem: str = None,
        max_moves: int = 10,
        num_rules: int = 0,
        problem_type: str = "mathy.unknown",
    ):
        self.parser = ExpressionParser()
        self.max_moves = max_moves
        self.num_rules = num_rules
        if problem is not None:
            self.agent = MathyAgentState(max_moves, problem, problem_type)
        elif state is not None:
            self.num_rules = state.num_rules
            self.max_moves = state.max_moves
            self.agent = MathyAgentState.copy(state.agent)
        else:
            raise ValueError("either state or a problem must be provided")
        # Ensure a string made it into the problem slot
        assert isinstance(self.agent.problem, str)

    @classmethod
    def copy(cls, from_state):
        return MathyEnvState(state=from_state)

    def clone(self):
        return MathyEnvState(state=self)

    def get_out_state(
        self, problem: str, action: int, focus_index: int, moves_remaining: int
    ) -> "MathyEnvState":
        """Get the next environment state based on the current one with updated
        history and agent information based on an action being taken."""
        out_state = MathyEnvState.copy(self)
        agent = out_state.agent
        agent.history.append(MathyEnvStateStep(problem, focus_index, action))
        agent.problem = problem
        agent.action = action
        agent.moves_remaining = moves_remaining
        agent.focus_index = focus_index
        return out_state

    def get_problem_hash(self) -> ProblemTypeIntList:
        """Return a two element array with hashed values for the current environment
        namespace string.

        # Example

        - `mycorp.envs.solve_impossible_problems` -> `[12375561, -2838517]`

        """
        # NOTE: import delayed to enable MP
        import tensorflow as tf

        global _problem_hash_cache
        if _problem_hash_cache is None:
            _problem_hash_cache = {}
        if self.agent.problem_type not in _problem_hash_cache:
            problem_hash_one = tf.strings.to_hash_bucket_strong(
                self.agent.problem_type, PROBLEM_TYPE_HASH_BUCKETS, [1337, 61059873]
            )
            problem_hash_two = tf.strings.to_hash_bucket_strong(
                self.agent.problem_type, PROBLEM_TYPE_HASH_BUCKETS, [7, -242315]
            )
            _problem_hash_cache[self.agent.problem_type] = [
                int(problem_hash_one.numpy()),
                int(problem_hash_two.numpy()),
            ]
        return _problem_hash_cache[self.agent.problem_type]

    def to_start_observation(
        self, rnn_state_h: RNNStateFloatList, rnn_state_c: RNNStateFloatList
    ) -> MathyObservation:
        """Generate an episode start MathyObservation"""
        num_actions = 1 * self.num_rules
        hash = self.get_problem_hash()
        mask = [0] * num_actions
        # HACKS: if you swap this line with below and train an agent on "poly,complex,binomial"
        #        it will crash after a few episodes, allowing you to test the error printing with
        #        call stack print outs.
        # values = [0.0] * num_actions
        values = [0.0]
        return MathyObservation(
            nodes=[MathTypeKeys["empty"]],
            mask=mask,
            values=values,
            type=hash,
            time=[-1.0],
            rnn_state_h=rnn_state_h,
            rnn_state_c=rnn_state_c,
            rnn_history_h=rnn_state_h,
        )

    def to_observation(
        self,
        move_mask: Optional[NodeMaskIntList] = None,
        rnn_state_h: Optional[RNNStateFloatList] = None,
        rnn_state_c: Optional[RNNStateFloatList] = None,
        rnn_history_h: Optional[RNNStateFloatList] = None,
        hash_type: Optional[ProblemTypeIntList] = None,
    ) -> MathyObservation:
        if rnn_state_h is None:
            rnn_state_h = rnn_placeholder_state(128)
        if rnn_state_c is None:
            rnn_state_c = rnn_placeholder_state(128)
        if rnn_history_h is None:
            rnn_history_h = rnn_placeholder_state(128)
        if hash_type is None:
            hash_type = self.get_problem_hash()
        expression = self.parser.parse(self.agent.problem)
        nodes: List[MathExpression] = expression.to_list()
        vectors: NodeIntList = []
        values: NodeValuesFloatList = []
        if move_mask is None:
            move_mask = np.zeros(len(nodes))
        for node in nodes:
            vectors.append(node.type_id)
            if isinstance(node, ConstantExpression):
                values.append(float(node.value))
            else:
                values.append(0.0)

        # Pass a 0-1 value indicating the relative episode time where 0.0 is
        # the episode start, and 1.0 is the episode end as indicated by the
        # maximum allowed number of actions.
        step = int(self.max_moves - self.agent.moves_remaining)
        time = step / self.max_moves

        return MathyObservation(
            nodes=vectors,
            mask=move_mask,
            values=values,
            type=hash_type,
            time=[time],
            rnn_state_h=rnn_state_h,
            rnn_state_c=rnn_state_c,
            rnn_history_h=rnn_history_h,
        )


class MathyAgentState:
    """The state related to an agent for a given environment state"""

    moves_remaining: int
    problem: str
    problem_type: str
    reward: float
    history: List[MathyEnvStateStep]
    focus_index: int
    last_action: int

    def __init__(
        self,
        moves_remaining,
        problem,
        problem_type,
        reward=0.0,
        history=None,
        focus_index=0,
        last_action=None,
    ):
        self.moves_remaining = moves_remaining
        self.focus_index = focus_index
        self.problem = problem
        self.last_action = last_action
        self.reward = reward
        self.problem_type = problem_type
        self.history = (
            history[:] if history is not None else [MathyEnvStateStep(problem, -1, -1)]
        )

    @classmethod
    def copy(cls, from_state):
        return MathyAgentState(
            moves_remaining=from_state.moves_remaining,
            problem=from_state.problem,
            reward=from_state.reward,
            problem_type=from_state.problem_type,
            history=from_state.history,
            focus_index=from_state.focus_index,
            last_action=from_state.last_action,
        )



from ..poly_simplify import PolySimplify
from ...types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register


class GymPolynomialSimplification(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs):
        super(GymPolynomialSimplification, self).__init__(
            env_class=PolySimplify,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialsEasy(GymPolynomialSimplification):
    def __init__(self, **kwargs):
        super(PolynomialsEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialsNormal(GymPolynomialSimplification):
    def __init__(self, **kwargs):
        super(PolynomialsNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialsHard(GymPolynomialSimplification):
    def __init__(self, **kwargs):
        super(PolynomialsHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(id="mathy-poly-easy-v0", entry_point="mathy.envs.gym:PolynomialsEasy")
safe_register(id="mathy-poly-normal-v0", entry_point="mathy.envs.gym:PolynomialsNormal")
safe_register(id="mathy-poly-hard-v0", entry_point="mathy.envs.gym:PolynomialsHard")



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tensorflow ops for multistep return evaluation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports
import tensorflow as tf


def _reverse_seq(sequence, sequence_lengths=None):
    """Reverse sequence along dim 0.

  Args:
    sequence: Tensor of shape [T, B, ...].
    sequence_lengths: (optional) tensor of shape [B]. If `None`, only reverse
      along dim 0.

  Returns:
    Tensor of same shape as sequence with dim 0 reversed up to sequence_lengths.
  """
    if sequence_lengths is None:
        return tf.reverse(sequence, [0])

    sequence_lengths = tf.convert_to_tensor(value=sequence_lengths)
    with tf.control_dependencies(
        [tf.compat.v1.assert_equal(sequence.shape[1], sequence_lengths.shape[0])]
    ):
        return tf.reverse_sequence(
            input=sequence, seq_lengths=sequence_lengths, seq_axis=0, batch_axis=1
        )


def scan_discounted_sum(
    sequence,
    decay,
    initial_value,
    reverse=False,
    sequence_lengths=None,
    back_prop=True,
    name="scan_discounted_sum",
):
    """Evaluates a cumulative discounted sum along dimension 0.

    ```python
    if reverse = False:
      result[1] = sequence[1] + decay[1] * initial_value
      result[k] = sequence[k] + decay[k] * result[k - 1]
    if reverse = True:
      result[last] = sequence[last] + decay[last] * initial_value
      result[k] = sequence[k] + decay[k] * result[k + 1]
    ```

  Respective dimensions T, B and ... have to be the same for all input tensors.
  T: temporal dimension of the sequence; B: batch dimension of the sequence.

    if sequence_lengths is set then x1 and x2 below are equivalent:
    ```python
    x1 = zero_pad_to_length(
      scan_discounted_sum(
          sequence[:length], decays[:length], **kwargs), length=T)
    x2 = scan_discounted_sum(sequence, decays,
                             sequence_lengths=[length], **kwargs)
    ```

  Args:
    sequence: Tensor of shape `[T, B, ...]` containing values to be summed.
    decay: Tensor of shape `[T, B, ...]` containing decays/discounts.
    initial_value: Tensor of shape `[B, ...]` containing initial value.
    reverse: Whether to process the sum in a reverse order.
    sequence_lengths: Tensor of shape `[B]` containing sequence lengths to be
      (reversed and then) summed.
    back_prop: Whether to backpropagate.
    name: Sets the name_scope for this op.

  Returns:
    Cumulative sum with discount. Same shape and type as `sequence`.
  """
    # Note this can be implemented in terms of cumprod and cumsum,
    # approximately as (ignoring boundary issues and initial_value):
    #
    # cumsum(decay_prods * sequence) / decay_prods
    # where decay_prods = reverse_cumprod(decay)
    #
    # One reason this hasn't been done is that multiplying then dividing again by
    # products of decays isn't ideal numerically, in particular if any of the
    # decays are zero it results in NaNs.
    with tf.compat.v1.name_scope(name, values=[sequence, decay, initial_value]):
        if sequence_lengths is not None:
            # Zero out sequence and decay beyond sequence_lengths.
            with tf.control_dependencies(
                [tf.compat.v1.assert_equal(sequence.shape[0], decay.shape[0])]
            ):
                mask = tf.sequence_mask(
                    sequence_lengths, maxlen=sequence.shape[0], dtype=sequence.dtype
                )
                mask = tf.transpose(a=mask)

            # Adding trailing dimensions to mask to allow for broadcasting.
            to_seq = mask.shape.dims + [1] * (sequence.shape.ndims - mask.shape.ndims)
            sequence *= tf.reshape(mask, to_seq)
            to_decay = mask.shape.dims + [1] * (decay.shape.ndims - mask.shape.ndims)
            decay *= tf.reshape(mask, to_decay)

        sequences = [sequence, decay]
        if reverse:
            sequences = [_reverse_seq(s, sequence_lengths) for s in sequences]

        summed = tf.scan(
            lambda a, x: x[0] + x[1] * a,
            sequences,
            initializer=tf.convert_to_tensor(value=initial_value),
            parallel_iterations=1,
            back_prop=back_prop,
        )
        if reverse:
            summed = _reverse_seq(summed, sequence_lengths)
        return summed


def multistep_forward_view(
    rewards,
    pcontinues,
    state_values,
    lambda_,
    back_prop=True,
    sequence_lengths=None,
    name="multistep_forward_view_op",
):
    """Evaluates complex backups (forward view of eligibility traces).

    ```python
    result[t] = rewards[t] +
        pcontinues[t]*(lambda_[t]*result[t+1] + (1-lambda_[t])*state_values[t])
    result[last] = rewards[last] + pcontinues[last]*state_values[last]
    ```

    This operation evaluates multistep returns where lambda_ parameter controls
    mixing between full returns and boostrapping. It is users responsibility
    to provide state_values. Depending on how state_values are evaluated this
    function can evaluate targets for Q(lambda), Sarsa(lambda) or some other
    multistep boostrapping algorithm.

    More information about a forward view is given here:
      http://incompleteideas.net/sutton/book/ebook/node74.html

    Please note that instead of evaluating traces and then explicitly summing
    them we instead evaluate mixed returns in the reverse temporal order
    by using the recurrent relationship given above.

    The parameter lambda_ can either be a constant value (e.g for Peng's
    Q(lambda) and Sarsa(_lambda)) or alternatively it can be a tensor containing
    arbitrary values (Watkins' Q(lambda), Munos' Retrace, etc).

    The result of evaluating this recurrence relation is a weighted sum of
    n-step returns, as depicted in the diagram below. One strategy to prove this
    equivalence notes that many of the terms in adjacent n-step returns
    "telescope", or cancel out, when the returns are summed.

    Below L3 is lambda at time step 3 (important: this diagram is 1-indexed, not
    0-indexed like Python). If lambda is scalar then L1=L2=...=Ln.
    g1,...,gn are discounts.

    ```
    Weights:  (1-L1)        (1-L2)*l1      (1-L3)*l1*l2  ...  L1*L2*...*L{n-1}
    Returns:    |r1*(g1)+     |r1*(g1)+      |r1*(g1)+          |r1*(g1)+
              v1*(g1)         |r2*(g1*g2)+   |r2*(g1*g2)+       |r2*(g1*g2)+
                            v2*(g1*g2)       |r3*(g1*g2*g3)+    |r3*(g1*g2*g3)+
                                           v3*(g1*g2*g3)               ...
                                                                |rn*(g1*...*gn)+
                                                              vn*(g1*...*gn)
    ```

  Args:
    rewards: Tensor of shape `[T, B]` containing rewards.
    pcontinues: Tensor of shape `[T, B]` containing discounts.
    state_values: Tensor of shape `[T, B]` containing state values.
    lambda_: Mixing parameter lambda.
        The parameter can either be a scalar or a Tensor of shape `[T, B]`
        if mixing is a function of state.
    back_prop: Whether to backpropagate.
    sequence_lengths: Tensor of shape `[B]` containing sequence lengths to be
      (reversed and then) summed, same as in `scan_discounted_sum`.
    name: Sets the name_scope for this op.

  Returns:
      Tensor of shape `[T, B]` containing multistep returns.
  """
    with tf.compat.v1.name_scope(name, values=[rewards, pcontinues, state_values]):
        # Regroup:
        #   result[t] = (rewards[t] + pcontinues[t]*(1-lambda_)*state_values[t]) +
        #               pcontinues[t]*lambda_*result[t + 1]
        # Define:
        #   sequence[t] = rewards[t] + pcontinues[t]*(1-lambda_)*state_values[t]
        #   discount[t] = pcontinues[t]*lambda_
        # Substitute:
        #   result[t] = sequence[t] + discount[t]*result[t + 1]
        # Boundary condition:
        #   result[last] = rewards[last] + pcontinues[last]*state_values[last]
        # Add and subtract the same quantity at BC:
        #   state_values[last] =
        #       lambda_*state_values[last] + (1-lambda_)*state_values[last]
        # This makes:
        #   result[last] =
        #       (rewards[last] + pcontinues[last]*(1-lambda_)*state_values[last]) +
        #       pcontinues[last]*lambda_*state_values[last]
        # Substitute in definitions for sequence and discount:
        #   result[last] = sequence[last] + discount[last]*state_values[last]
        # Define:
        #   initial_value=state_values[last]
        # We get the following recurrent relationship:
        #   result[last] = sequence[last] + decay[last]*initial_value
        #   result[k] = sequence[k] + decay[k] * result[k + 1]
        # This matches the form of scan_discounted_sum:
        #   result = scan_sum_with_discount(sequence, discount,
        #                                   initial_value = state_values[last])
        sequence = rewards + pcontinues * state_values * (1 - lambda_)
        discount = pcontinues * lambda_
        return scan_discounted_sum(
            sequence,
            discount,
            state_values[-1],
            reverse=True,
            sequence_lengths=sequence_lengths,
            back_prop=back_prop,
        )




import math
import os
import queue
import threading
import time
from multiprocessing import Queue
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import gym
import numpy as np

from ...envs.gym.mathy_gym_env import MathyGymEnv
from ...util import calculate_grouping_control_signal
from ...state import (
    MathyEnvState,
    MathyObservation,
    MathyWindowObservation,
    observations_to_window,
)
from ...teacher import Teacher
from ...util import discount
from .. import action_selectors
from ..episode_memory import EpisodeMemory
from ..policy_value_model import PolicyValueModel
from ..trfl import discrete_policy_entropy_loss, td_lambda
from .config import SelfPlayConfig
from .lib.average_meter import AverageMeter
from .lib.progress.bar import Bar


class SelfPlayTrainer:

    args: SelfPlayConfig

    def __init__(
        self, args: SelfPlayConfig, model: PolicyValueModel, action_size: int,
    ):
        super(SelfPlayTrainer, self).__init__()
        import tensorflow as tf

        self.args = args
        self.model = model
        self.iteration = 0
        self.action_size = action_size
        self.writer = None
        self.last_histogram_write = -1
        if self.args.use_grouping_control:
            raise NotImplementedError(
                "Grouping Control signal is not implemented for the Zero agent."
                " Support shouldn't be very difficult to add by looking at the"
                " compute_policy_value_loss in the A3C agent."
            )

    @property
    def tb_prefix(self) -> str:
        return "agent"

    def train(self, examples, model):

        total_batches = int(len(examples) / self.args.batch_size)
        if total_batches == 0:
            return False

        import tensorflow as tf

        print(
            "Training neural net for ({}) epochs with ({}) examples...".format(
                self.args.epochs, len(examples)
            )
        )

        for epoch in range(self.args.epochs):
            print("EPOCH ::: " + str(epoch + 1))
            data_time = AverageMeter()
            batch_time = AverageMeter()
            pi_losses = AverageMeter()
            v_losses = AverageMeter()
            end = time.time()

            batch_steps = int(len(examples) / self.args.batch_size)
            bar = Bar("Training Net", max=batch_steps)
            batch_idx = 0

            # self.session.run(tf.local_variables_initializer())
            while batch_idx < total_batches:
                sample_ids = np.random.randint(len(examples), size=self.args.batch_size)
                (
                    text,
                    action,
                    reward,
                    discounted,
                    terminal,
                    observation,
                    pi_,
                    v,
                ) = list(zip(*[examples[i] for i in sample_ids]))
                pi = tf.keras.preprocessing.sequence.pad_sequences(pi_)
                inputs = observations_to_window(
                    [MathyObservation(*o) for o in observation]
                )
                with tf.GradientTape() as tape:
                    pi_loss, value_loss, total_loss = self.compute_loss(
                        gamma=self.args.gamma, inputs=inputs, target_pi=pi, target_v=v,
                    )
                grads = tape.gradient(total_loss, self.model.trainable_weights)
                zipped_gradients = zip(grads, self.model.trainable_weights)
                self.model.optimizer.apply_gradients(zipped_gradients)

                # measure data loading time
                data_time.update(time.time() - end)

                pi_losses.update(pi_loss, len(text))
                v_losses.update(value_loss, len(text))

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()
                batch_idx += 1

                # plot progress
                bar.suffix = "({batch}/{size}) data: {data:.3f}s batch: {bt:.3f}s total: {total:} eta: {eta:} pi: {lpi:.4f} v: {lv:.3f}".format(
                    batch=batch_idx,
                    size=batch_steps,
                    data=data_time.avg,
                    bt=batch_time.avg,
                    total=bar.elapsed_td,
                    eta=bar.eta_td,
                    lpi=pi_losses.avg,
                    lv=v_losses.avg,
                )

                bar.next()
            bar.finish()

        return True

    def compute_policy_value_loss(
        self, inputs: MathyWindowObservation, target_pi: Any, target_v: Any, gamma=0.99,
    ):
        import tensorflow as tf

        batch_size = len(inputs.nodes)
        step = self.model.optimizer.iterations
        logits, values, _ = self.model(inputs.to_inputs())
        value_loss = tf.losses.mean_squared_error(
            target_v, tf.reshape(values, shape=[-1])
        )
        policy_logits = tf.reshape(logits, [batch_size, -1])
        policy_logits = policy_logits[:, : target_pi.shape[1]]
        policy_loss = tf.nn.softmax_cross_entropy_with_logits(
            labels=target_pi, logits=policy_logits
        )
        policy_loss = tf.reduce_mean(policy_loss)
        total_loss = value_loss + policy_loss
        prefix = self.tb_prefix
        tf.summary.scalar(f"{prefix}/policy_loss", data=policy_loss, step=step)
        tf.summary.scalar(f"{prefix}/value_loss", data=value_loss, step=step)

        return (policy_loss, value_loss, total_loss)

    def compute_loss(
        self, inputs: MathyWindowObservation, target_pi: Any, target_v: Any, gamma=0.99,
    ):
        loss_tuple = self.compute_policy_value_loss(inputs, target_pi, target_v)
        pi_loss, v_loss, total_loss = loss_tuple
        return pi_loss, v_loss, total_loss



from typing import Any, List, Optional, Type

from numpy.random import randint

from .. import time_step
from ..core.expressions import MathExpression
from ..util import TermEx, get_term_ex, get_terms
from ..env import MathyEnv, MathyEnvProblem
from ..problems import gen_commute_haystack
from ..core.rule import BaseRule
from ..state import MathyEnvState, MathyObservation
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs


class PolyGroupLikeTerms(MathyEnv):
    """A Mathy environment for grouping polynomial terms that are like.

    The goal is to commute all the like terms so they become siblings as quickly as
    possible.
    """

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.group_like_terms"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity * 2

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If all like terms are siblings."""
        term_nodes = get_terms(expression)
        already_seen: set = set()
        current_term = ""
        # Iterate over each term in order and build a unique key to identify its
        # term likeness. For this we drop the coefficient from the term and use
        # only its variable/exponent to build keys.
        for term in term_nodes:
            ex: Optional[TermEx] = get_term_ex(term)
            if ex is None:
                raise ValueError("should this happen?")
            key = f"{ex.variable}{ex.exponent}"
            # If the key is in the "already seen and moved on" list then we've failed
            # to meet the completion criteria. e.g. the final x in "4x + 2y + x"
            if key in already_seen:
                return None
            if key != current_term:
                already_seen.add(current_term)
                current_term = key

        return time_step.termination(features, self.get_win_signal(env_state))

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        if params.difficulty == MathyEnvDifficulty.easy:
            blockers = randint(1, 3)
            text, _ = gen_commute_haystack(
                commute_blockers=1, min_terms=5, max_terms=7, easy=True, powers=False
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            blockers = randint(2, 4)
            text, _ = gen_commute_haystack(
                min_terms=5, max_terms=10, easy=False, powers=True
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            blockers = randint(3, 6)
            text, _ = gen_commute_haystack(
                min_terms=11, max_terms=16, easy=False, powers=True
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, blockers + 2, self.get_env_namespace())



from ..core.expressions import (
    AddExpression,
    MultiplyExpression,
    ConstantExpression,
    VariableExpression,
    PowerExpression,
    MathExpression,
)
from ..core.rule import BaseRule


class CommutativeSwapRule(BaseRule):
    r"""Commutative Property
    For Addition: `a + b = b + a`

             +                  +
            / \                / \
           /   \     ->       /   \
          /     \            /     \
         a       b          b       a

    For Multiplication: `a * b = b * a`

             *                  *
            / \                / \
           /   \     ->       /   \
          /     \            /     \
         a       b          b       a
    """
    preferred: bool

    def __init__(self, preferred=True):
        # If false, terms that are in preferred order will not commute
        self.preferred = preferred

    @property
    def name(self):
        return "Commutative Swap"

    @property
    def code(self):
        return "CS"

    def can_apply_to(self, node: MathExpression):
        # Must be an add/multiply
        if isinstance(node, AddExpression):
            return True
        if not isinstance(node, MultiplyExpression):
            return False
        # When preferred is false, the commutative rule won't apply to term
        # nodes that are already in a preferred position.
        if self.preferred is False:
            # 4x won't commute to x * 4
            left_const = isinstance(node.left, ConstantExpression)
            right_var = isinstance(node.right, VariableExpression)
            if left_const and right_var:
                # UNLESS it is within a more complex term that spans multiple nodes.
                if node.parent and isinstance(node.parent, MultiplyExpression):
                    sibling = node.get_sibling()
                    return sibling and isinstance(sibling, MultiplyExpression)
                # Nope
                return False

            # 8y^4 won't commute to y^4 * 8
            if isinstance(node.right, PowerExpression):
                right_left_var = isinstance(node.right.left, VariableExpression)
                right_right_const = isinstance(node.right.right, ConstantExpression)
                if right_left_var and right_right_const:
                    # UNLESS it is within a more complex term that spans multiple nodes.
                    if node.parent and isinstance(node.parent, MultiplyExpression):
                        sibling = node.get_sibling()
                        return sibling and isinstance(sibling, MultiplyExpression)
                    # Nope
                    return False
        return True

    def apply_to(self, node: MathExpression):
        change = super().apply_to(node)
        a = node.left
        b = node.right
        assert a is not None

        add_chain = isinstance(a, AddExpression) and isinstance(node, AddExpression)
        mul_chain = isinstance(a, MultiplyExpression) and isinstance(
            node, MultiplyExpression
        )
        # The left node is not another sibling add
        if not add_chain and not mul_chain:
            node.set_right(a)
            node.set_left(b)
        else:
            # If it's another add, just swap their
            # children directly to avoid inner-nesting.
            one = a.left
            two = a.right
            three = node.right
            a.set_left(three)
            a.set_right(one)
            node.set_right(two)

        node.set_changed()
        change.done(node)
        return change



from .mathy_gym_env import *  # noqa
from .masked_discrete import *  # noqa
from .gym_binomial_distribute import *  # noqa
from .gym_complex_simplify import *  # noqa
from .gym_poly_blockers import *  # noqa
from .gym_poly_grouping import *  # noqa
from .gym_poly_haystack_like_terms import *  # noqa
from .gym_poly_simplify import *  # noqa
from .gym_poly_combine_in_place import *  # noqa
from .gym_poly_commute_like_terms import *  # noqa



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Deterministic Policy Gradient (DPG) ops.

These ops support training a value based agent on control problems with
continuous action spaces. The agent's actions are assumed to be continuous
vectors of size `action_dimension`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
import tensorflow as tf
from . import base_ops

DPGExtra = collections.namedtuple("dpg_extra", ["q_max", "a_max", "dqda"])


def dpg(q_max, a_max, dqda_clipping=None, clip_norm=False, name="DpgLearning"):
    """Implements the Deterministic Policy Gradient (DPG) loss as a TensorFlow Op.

  This op implements the loss for the `actor`, the `critic` can instead be
  updated by minimizing the `value_ops.td_learning` loss.

  See "Deterministic Policy Gradient Algorithms" by Silver, Lever, Heess,
  Degris, Wierstra, Riedmiller (http://proceedings.mlr.press/v32/silver14.pdf).

  Args:
    q_max: Tensor holding Q-values generated by Q network with the input of
      (state, a_max) pair, shape `[B]`.
    a_max: Tensor holding the optimal action, shape `[B, action_dimension]`.
    dqda_clipping: `int` or `float`, clips the gradient dqda element-wise
      between `[-dqda_clipping, dqda_clipping]`.
    clip_norm: Whether to perform dqda clipping on the vector norm of the last
      dimension, or component wise (default).
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `q_max`: Tensor holding the optimal Q values, `[B]`.
        * `a_max`: Tensor holding the optimal action, `[B, action_dimension]`.
        * `dqda`: Tensor holding the derivative dq/da, `[B, action_dimension]`.

  Raises:
    ValueError: If `q_max` doesn't depend on `a_max` or if `dqda_clipping <= 0`.
  """

    # DPG op.
    with tf.compat.v1.name_scope(name, values=[q_max, a_max]):

        # Calculate the gradient dq/da.
        dqda = tf.gradients(ys=[q_max], xs=[a_max])[0]

        # Check that `q_max` depends on `a_max`.
        if dqda is None:
            raise ValueError("q_max needs to be a function of a_max")

        # Clipping the gradient dq/da.
        if dqda_clipping is not None:
            if dqda_clipping <= 0:
                raise ValueError(
                    "dqda_clipping should be bigger than 0, {} found".format(
                        dqda_clipping
                    )
                )
            if clip_norm:
                dqda = tf.clip_by_norm(dqda, dqda_clipping, axes=-1)
            else:
                dqda = tf.clip_by_value(dqda, -1.0 * dqda_clipping, dqda_clipping)

        # Target_a ensures correct gradient calculated during backprop.
        target_a = dqda + a_max
        # Stop the gradient going through Q network when backprop.
        target_a = tf.stop_gradient(target_a)
        # Gradient only go through actor network.
        loss = 0.5 * tf.reduce_sum(input_tensor=tf.square(target_a - a_max), axis=-1)
        return base_ops.LossOutput(loss, DPGExtra(q_max=q_max, a_max=a_max, dqda=dqda))



import json
from pathlib import Path
from typing import Type

from .core.parser import ExpressionParser
from .core.rule import BaseRule
from .util import compare_expression_string_values, compare_expression_values
from .rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
    VariableMultiplyRule,
)


def get_rule_tests(name):
    """Load a set of JSON rule test assertions.

    # Arguments
    name (str): The name of the test JSON file to open, e.g. "commutative_property"

    # Returns
    (dict): A dictionary with "valid" and "invalid" keys that contain pairs of 
    expected inputs and outputs.
    """
    rule_file = (
        Path(__file__).parent.parent / "tests" / "rules" / "{}.json".format(name)
    )
    if not rule_file.is_file() is True:
        raise ValueError(f"does not exist: {rule_file}")
    with open(rule_file, "r") as file:
        return json.load(file)


def init_rule_for_test(example: dict, rule_class: Type[BaseRule]) -> BaseRule:
    """Initialize a given rule_class from a test example.

    This handles optionally passing the test example constructor arguments
    to the Rule.

    # Arguments:
    example (dict): The example assertion loaded from a call to `get_rule_tests`
    rule_class (Type[BaseRule]): The 

    # Returns
    (BaseRule): The rule instance.
    """
    if "args" not in example:
        rule = rule_class()
    else:
        rule = rule_class(**example["args"])  # type: ignore
    return rule


def run_rule_tests(name, rule_class, callback=None):
    """Load and assert about the transformations and validity of rules
    based on given input examples.

    When debugging a problem it can be useful to provide a "callback" function
    and add a `"debug": true` value to the example in the rules json file you
    want to debug. Then you set a breakpoint and step out of your callback function
    into the parsing/evaluation of the debug example.
    """
    tests = get_rule_tests(name)
    parser = ExpressionParser()
    for ex in tests["valid"]:
        # Trigger the debug callback so the user can step over into the useful stuff
        if callback is not None:
            callback(ex)
        rule = init_rule_for_test(ex, rule_class)
        expression = parser.parse(ex["input"]).clone()
        before = expression.clone()
        print(ex)
        if "target" in ex:
            nodes = rule.find_nodes(expression)
            targets = [n.raw for n in nodes]
            node = [n for n in nodes if n.raw == ex["target"]]
            assert len(node) > 0, f"could not find target node. targets are: {targets}"
            node = node[0]
        else:
            node = rule.find_node(expression)

        if node is None:
            assert "expected to find node but did not for" == str(expression)
        change = rule.apply_to(node)
        after = change.result.get_root()
        # Compare the values of the in-memory expressions output from the rule
        compare_expression_values(before, after)
        # Parse the output strings to new expressions, and compare the values
        compare_expression_string_values(str(before), str(after))
        assert str(after).strip() == ex["output"]

    for ex in tests["invalid"]:
        # Trigger the debug callback so the user can step over into the useful stuff
        if callback is not None:
            callback(ex)
        rule = init_rule_for_test(ex, rule_class)
        expression = parser.parse(ex["input"]).clone()
        node = rule.find_node(expression)
        if node is not None:
            raise ValueError(
                "expected not to find a node, but found: {}".format(str(node))
            )



import json
import time
from pathlib import Path

import numpy as np

from mathy.core.parser import ExpressionParser, ParserException

from ...env import MathyEnv
from .config import SelfPlayConfig
from .lib.average_meter import AverageMeter
from .lib.progress.bar import Bar
from .practice_runner import ParallelPracticeRunner, PracticeRunner
from .practice_session import PracticeSession


def self_play_runner(config: SelfPlayConfig):
    """Practice a concept for up to (n) lessons or until the concept is learned as defined
    by the lesson plan. """
    BaseEpisodeRunner = (
        PracticeRunner if config.num_workers < 2 else ParallelPracticeRunner
    )
    if config.profile and config.num_workers > 1:
        raise NotImplementedError("zero agent does not support multiprocess profiling")
    lesson_name = "mathy-poly-easy-v0"
    if config.verbose:
        print(config.json(indent=2))

    class LessonRunner(BaseEpisodeRunner):  # type:ignore
        def get_env(self):
            import gym

            return gym.make(lesson_name)

        def get_model(self, game):
            from ...agents.policy_value_model import (
                get_or_create_policy_model,
                PolicyValueModel,
            )

            model: PolicyValueModel = get_or_create_policy_model(
                args=config, predictions=game.action_space.n, is_main=True
            )
            return model

    print("Practicing {}...".format(config.topics))
    runner = LessonRunner(config)
    c = PracticeSession(runner, config, env_name=lesson_name)
    c.learn()



from ..poly_commute_like_terms import PolyCommuteLikeTerms
from ...types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register

#
# Combine like terms without commuting
#


class GymPolynomialCommuteLikeTerms(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs):
        super(GymPolynomialCommuteLikeTerms, self).__init__(
            env_class=PolyCommuteLikeTerms,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialCommuteLikeTermsEasy(GymPolynomialCommuteLikeTerms):
    def __init__(self, **kwargs):
        super(PolynomialCommuteLikeTermsEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialCommuteLikeTermsNormal(GymPolynomialCommuteLikeTerms):
    def __init__(self, **kwargs):
        super(PolynomialCommuteLikeTermsNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialCommuteLikeTermsHard(GymPolynomialCommuteLikeTerms):
    def __init__(self, **kwargs):
        super(PolynomialCommuteLikeTermsHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(
    id="mathy-poly-commute-easy-v0",
    entry_point="mathy.envs.gym:PolynomialCommuteLikeTermsEasy",
)
safe_register(
    id="mathy-poly-commute-normal-v0",
    entry_point="mathy.envs.gym:PolynomialCommuteLikeTermsNormal",
)
safe_register(
    id="mathy-poly-commute-hard-v0",
    entry_point="mathy.envs.gym:PolynomialCommuteLikeTermsHard",
)
